#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import json
import os
import re
from typing import Any, Optional

import jq  # type: ignore

from . import credentials
from .credentials import CredentialProvider
from .credentials.bitwarden_provider import BitwardenVault
from .credentials.inmemory_provider import InmemoryVault
from .credentials.pass_provider import PassVault
from .jsonnet import JsonnetConfig


class OrganizationConfig:
    def __init__(
        self,
        name: str,
        github_id: str,
        config_repo: str,
        base_template: str,
        jsonnet_config: JsonnetConfig,
        credential_data: dict[str, Any],
    ):
        self._name = name
        self._github_id = github_id
        self._config_repo = config_repo
        self._base_template = base_template
        self._jsonnet_config = jsonnet_config
        self._credential_data = credential_data

    @property
    def name(self):
        return self._name

    @property
    def github_id(self) -> str:
        return self._github_id

    @property
    def config_repo(self) -> str:
        return self._config_repo

    @property
    def base_template(self) -> str:
        return self._base_template

    @property
    def jsonnet_config(self) -> JsonnetConfig:
        return self._jsonnet_config

    @property
    def credential_data(self) -> dict[str, Any]:
        return self._credential_data

    @credential_data.setter
    def credential_data(self, data: dict[str, Any]) -> None:
        self._credential_data = data

    def __repr__(self) -> str:
        return (
            f"OrganizationConfig('{self.name}', '{self.github_id}', "
            f"'{self.config_repo}', {json.dumps(self.credential_data)})"
        )

    @classmethod
    def from_dict(cls, data: dict[str, Any], otterdog_config: OtterdogConfig) -> OrganizationConfig:
        name = data.get("name")
        if name is None:
            raise RuntimeError(f"missing required name for organization config with data: '{json.dumps(data)}'")

        github_id = data.get("github_id")
        if github_id is None:
            raise RuntimeError(f"missing required github_id for organization config with name '{name}'")

        config_repo = data.get("config_repo", otterdog_config.default_config_repo)
        base_template = data.get("base_template", otterdog_config.default_base_template)

        jsonnet_config = JsonnetConfig(
            github_id,
            otterdog_config.jsonnet_base_dir,
            base_template,
            otterdog_config.local_mode,
        )

        data = data.get("credentials", {})
        if data is None:
            raise RuntimeError(f"missing required credentials for organization config with name '{name}'")

        return cls(name, github_id, config_repo, base_template, jsonnet_config, data)

    @classmethod
    def of(
        cls,
        project_name: str,
        github_id: str,
        credential_data: dict[str, Any],
        work_dir: str,
        otterdog_config: OtterdogConfig,
    ) -> OrganizationConfig:
        # get the default organization config to retrieve some properties
        org_config = otterdog_config.get_organization_config(project_name)
        config_repo = org_config.config_repo
        base_template = org_config.base_template

        # construct a custom jsonnet config based on the given work_dir
        base_dir = os.path.join(otterdog_config.jsonnet_base_dir, work_dir)

        jsonnet_config = JsonnetConfig(
            github_id,
            base_dir,
            base_template,
            otterdog_config.local_mode,
        )

        return cls(project_name, github_id, config_repo, base_template, jsonnet_config, credential_data)


class OtterdogConfig:
    def __init__(self, config_file: str, local_mode: bool, working_dir: Optional[str] = None):
        if not os.path.exists(config_file):
            raise RuntimeError(f"configuration file '{config_file}' not found")

        self._config_file = os.path.realpath(config_file)
        self._config_dir = os.path.dirname(self._config_file)
        self._credential_providers: dict[str, CredentialProvider] = {}

        self._local_mode = local_mode

        with open(config_file) as f:
            self._configuration = json.load(f)

        self._jsonnet_config = jq.compile(".defaults.jsonnet // {}").input(self._configuration).first()
        self._github_config = jq.compile(".defaults.github // {}").input(self._configuration).first()
        self._default_credential_provider = (
            jq.compile('.defaults.credentials.provider // ""').input(self._configuration).first()
        )

        if working_dir is None:
            self._jsonnet_base_dir = os.path.join(self._config_dir, self._jsonnet_config.get("config_dir", "orgs"))
        else:
            self._jsonnet_base_dir = os.path.join(working_dir, self._jsonnet_config.get("config_dir", "orgs"))
            if not os.path.exists(self._jsonnet_base_dir):
                os.makedirs(self._jsonnet_base_dir)

        organizations = self._configuration.get("organizations", [])

        self._organizations_map = {}
        self._organizations = []
        for org in organizations:
            org_config = OrganizationConfig.from_dict(org, self)
            self._organizations.append(org_config)
            self._organizations_map[org_config.name] = org_config
            self._organizations_map[org_config.github_id] = org_config

    @property
    def config_file(self) -> str:
        return self._config_file

    @property
    def jsonnet_base_dir(self) -> str:
        return self._jsonnet_base_dir

    @property
    def local_mode(self) -> bool:
        return self._local_mode

    @property
    def default_config_repo(self) -> str:
        return self._github_config.get("config_repo", ".otterdog")

    @property
    def default_base_template(self) -> str:
        return self._jsonnet_config["base_template"]

    @property
    def organization_configs(self) -> dict[str, OrganizationConfig]:
        return self._organizations_map

    @property
    def organization_names(self) -> list[str]:
        return list(map(lambda config: config.name, self._organizations))

    def get_project_name(self, github_id: str) -> Optional[str]:
        organization = self._organizations_map.get(github_id)
        if organization is not None:
            return organization.name
        else:
            return None

    def get_organization_config(self, organization_name: str) -> OrganizationConfig:
        org_config = self._organizations_map.get(organization_name)
        if org_config is None:
            raise RuntimeError(f"unknown organization with name / github_id '{organization_name}'")
        return org_config

    def _get_credential_provider(self, provider_type: str) -> credentials.CredentialProvider:
        provider = self._credential_providers.get(provider_type)
        if provider is None:
            match provider_type:
                case "bitwarden":
                    api_token_key = (
                        jq.compile('.defaults.bitwarden.api_token_key // "api_token_admin"')
                        .input(self._configuration)
                        .first()
                    )

                    provider = BitwardenVault(api_token_key)
                    self._credential_providers[provider_type] = provider

                case "pass":
                    password_store_dir = (
                        jq.compile('.defaults.pass.password_store_dir // ""').input(self._configuration).first()
                    )

                    username_pattern = (
                        jq.compile('.defaults.pass.username_pattern // ""').input(self._configuration).first()
                    )

                    password_pattern = (
                        jq.compile('.defaults.pass.password_pattern // ""').input(self._configuration).first()
                    )

                    twofa_seed_pattern = (
                        jq.compile('.defaults.pass.twofa_seed_pattern // ""').input(self._configuration).first()
                    )

                    api_token_pattern = (
                        jq.compile('.defaults.pass.api_token_pattern // ""').input(self._configuration).first()
                    )

                    provider = PassVault(
                        password_store_dir, username_pattern, password_pattern, twofa_seed_pattern, api_token_pattern
                    )
                    self._credential_providers[provider_type] = provider

                case "inmemory":
                    provider = InmemoryVault()
                    self._credential_providers[provider_type] = provider

                case _:
                    raise RuntimeError(f"unsupported credential provider '{provider_type}'")

        return provider

    def get_credentials(self, org_config: OrganizationConfig, only_token: bool = False) -> credentials.Credentials:
        provider_type = org_config.credential_data.get("provider")

        if provider_type is None:
            provider_type = self._default_credential_provider

        if not provider_type:
            raise RuntimeError(f"no credential provider configured for organization '{org_config.name}'")

        provider = self._get_credential_provider(provider_type)
        return provider.get_credentials(org_config.name, org_config.credential_data, only_token)

    def get_secret(self, secret_data: str) -> str:
        if secret_data and ":" in secret_data:
            provider_type, data = re.split(":", secret_data)
            provider = self._get_credential_provider(provider_type)
            return provider.get_secret(data)
        else:
            return secret_data

    def __repr__(self):
        return f"OtterdogConfig('{self.config_file}')"

    @classmethod
    def from_file(cls, config_file: str, local_mode: bool):
        return cls(config_file, local_mode)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from datetime import datetime
from typing import Any

from otterdog.providers.github.exception import GitHubException
from otterdog.utils import print_debug

from . import RestApi, RestClient, parse_date_string


class AppClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_authenticated_app(self) -> dict[str, Any]:
        print_debug("retrieving authenticated app")

        try:
            return await self.requester.async_request_json("GET", "/app")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving authenticated app:\n{ex}").with_traceback(tb)

    async def get_app_installations(self) -> list[dict[str, Any]]:
        print_debug("retrieving app installations")

        try:
            return await self.requester.async_request_paged_json("GET", "/app/installations")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving authenticated app:\n{ex}").with_traceback(tb)

    async def create_installation_access_token(self, installation_id: str) -> tuple[str, datetime]:
        print_debug(f"creating an installation access token for installation '{installation_id}'")

        try:
            response = await self.requester.async_request_json(
                "POST", f"/app/installations/{installation_id}/access_tokens"
            )
            return response["token"], parse_date_string(response["expires_at"])
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed creating installation access token:\n{ex}").with_traceback(tb)

    async def get_app_ids(self, app_slug: str) -> tuple[int, str]:
        print_debug("retrieving app node id")

        try:
            response = await self.requester.async_request_json("GET", f"/apps/{app_slug}")
            return response["id"], response["node_id"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving app node id:\n{ex}").with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import associate_by_key, style

from . import Operation


class DeleteFileOperation(Operation):
    """
    Deletes a file identified by its path in an organization repository.
    """

    def __init__(self, repo: str, path: str, message: str):
        super().__init__()
        self._repo = repo
        self._path = path
        self._message = message

    @property
    def repo(self) -> str:
        return self._repo

    @property
    def path(self) -> str:
        return self._path

    @property
    def message(self) -> str:
        return self._message

    def pre_execute(self) -> None:
        self.printer.println(f"Deleting file '{self._path}' in organization repository '{self.repo}':")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file

            if not os.path.exists(org_file_name):
                self.printer.print_error(
                    f"configuration file '{org_file_name}' does not yet exist, run 'fetch-config' or 'import' first."
                )
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
            except RuntimeError as ex:
                self.printer.print_error(f"failed to load configuration: {str(ex)}")
                return 1

            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                rest_api = provider.rest_api

                repositories_by_name = associate_by_key(organization.repositories, lambda r: r.name)
                repo = repositories_by_name.get(self.repo)
                if repo is not None and repo.archived is False:
                    collected_error = None
                    repo_name = f"{github_id}/{repo.name}"
                    self.printer.print(
                        f"Deleting file '{style(self.path, bright=True)}' "
                        f"in repository '{style(repo_name, bright=True)}': "
                    )

                    try:
                        deleted_file = await rest_api.content.delete_content(
                            github_id, repo.name, self.path, self.message
                        )
                    except RuntimeError as e:
                        collected_error = e

                    if deleted_file is True:
                        self.printer.println(style("succeeded", fg="green"))
                    else:
                        self.printer.println(style("failed", fg="red"))

                    if collected_error is not None:
                        self.printer.println()
                        self.printer.print_error(f"failure deleting file\n{str(collected_error)}")
                        return 1
            return 0

        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any, Optional

from importlib_resources import files

from otterdog import resources, utils
from otterdog.credentials import Credentials
from otterdog.providers.github.auth import token_auth

from .graphql import GraphQLClient
from .rest import RestApi
from .web import WebClient

_ORG_SETTINGS_SCHEMA = json.loads(files(resources).joinpath("schemas/settings.json").read_text())


class GitHubProvider:
    def __init__(self, credentials: Optional[Credentials]):
        self._credentials = credentials

        self._settings_schema = _ORG_SETTINGS_SCHEMA
        # collect supported rest api keys
        self._settings_restapi_keys = {
            k for k, v in self._settings_schema["properties"].items() if v.get("provider") == "restapi"
        }

        # collect supported web interface keys
        self._settings_web_keys = {
            k for k, v in self._settings_schema["properties"].items() if v.get("provider") == "web"
        }
        # TODO: make this cleaner
        self._settings_web_keys.add("discussion_source_repository_id")

        if credentials is not None:
            self._init_clients()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def close(self) -> None:
        if self._credentials is not None:
            self.rest_api.close()

    def _init_clients(self):
        self.rest_api = RestApi(token_auth(self._credentials.github_token))
        self.web_client = WebClient(self._credentials)
        self.graphql_client = GraphQLClient(token_auth(self._credentials.github_token))

    def __getstate__(self):
        return (
            self._credentials,
            self._settings_schema,
            self._settings_restapi_keys,
            self._settings_web_keys,
        )

    def __setstate__(self, state):
        (
            self._credentials,
            self._settings_schema,
            self._settings_restapi_keys,
            self._settings_web_keys,
        ) = state
        self._init_clients()

    async def get_content(self, org_id: str, repo_name: str, path: str, ref: Optional[str] = None) -> str:
        return await self.rest_api.content.get_content(org_id, repo_name, path, ref)

    async def update_content(
        self,
        org_id: str,
        repo_name: str,
        path: str,
        content: str,
        message: Optional[str] = None,
    ) -> bool:
        return await self.rest_api.content.update_content(org_id, repo_name, path, content, message)

    async def get_org_settings(self, org_id: str, included_keys: set[str], no_web_ui: bool) -> dict[str, Any]:
        # first, get supported settings via the rest api.
        required_rest_keys = {x for x in included_keys if x in self._settings_restapi_keys}
        merged_settings = await self.rest_api.org.get_settings(org_id, required_rest_keys)

        # second, get settings only accessible via the web interface and merge
        # them with the other settings, unless --no-web-ui is specified.
        if not no_web_ui:
            required_web_keys = {x for x in included_keys if x in self._settings_web_keys}
            if len(required_web_keys) > 0:
                web_settings = await self.web_client.get_org_settings(org_id, required_web_keys)
                merged_settings.update(web_settings)

            utils.print_trace(f"merged org settings = {merged_settings}")

        return merged_settings

    async def update_org_settings(self, org_id: str, settings: dict[str, Any]) -> None:
        rest_fields = {}
        web_fields = {}

        # split up settings to be updated whether they need be updated
        # via rest api or web interface.
        for k, v in sorted(settings.items()):
            if k in self._settings_restapi_keys:
                rest_fields[k] = v
            elif k in self._settings_web_keys:
                web_fields[k] = v
            else:
                utils.print_warn(f"encountered unknown field '{k}' during update, ignoring")

        # update any settings via the rest api
        if len(rest_fields) > 0:
            await self.rest_api.org.update_settings(org_id, rest_fields)

        # update any settings via the web interface
        if len(web_fields) > 0:
            await self.web_client.update_org_settings(org_id, web_fields)

    async def get_org_workflow_settings(self, org_id: str) -> dict[str, Any]:
        return await self.rest_api.org.get_workflow_settings(org_id)

    async def update_org_workflow_settings(self, org_id: str, workflow_settings: dict[str, Any]) -> None:
        await self.rest_api.org.update_workflow_settings(org_id, workflow_settings)

    async def get_org_webhooks(self, org_id: str) -> list[dict[str, Any]]:
        return await self.rest_api.org.get_webhooks(org_id)

    async def update_org_webhook(self, org_id: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        if len(webhook) > 0:
            await self.rest_api.org.update_webhook(org_id, webhook_id, webhook)

    async def add_org_webhook(self, org_id: str, data: dict[str, str]) -> None:
        await self.rest_api.org.add_webhook(org_id, data)

    async def delete_org_webhook(self, org_id: str, webhook_id: int, url: str) -> None:
        await self.rest_api.org.delete_webhook(org_id, webhook_id, url)

    async def get_repos(self, org_id: str) -> list[str]:
        # filter out repos which are created to work on GitHub Security Advisories
        # they should not be part of the visible configuration
        return list(filter(lambda name: not utils.is_ghsa_repo(name), await self.rest_api.org.get_repos(org_id)))

    async def get_repo_data(self, org_id: str, repo_name: str) -> dict[str, Any]:
        return await self.rest_api.repo.get_repo_data(org_id, repo_name)

    async def get_repo_by_id(self, repo_id: int) -> dict[str, Any]:
        return await self.rest_api.repo.get_repo_by_id(repo_id)

    async def update_repo(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        if len(data) > 0:
            await self.rest_api.repo.update_repo(org_id, repo_name, data)

    async def add_repo(
        self,
        org_id: str,
        data: dict[str, str],
        template_repository: Optional[str],
        post_process_template_content: list[str],
        forked_repository: Optional[str],
        fork_default_branch_only: bool,
        auto_init_repo: bool,
    ) -> None:
        await self.rest_api.repo.add_repo(
            org_id,
            data,
            template_repository,
            post_process_template_content,
            forked_repository,
            fork_default_branch_only,
            auto_init_repo,
        )

    async def delete_repo(self, org_id: str, repo_name: str) -> None:
        await self.rest_api.repo.delete_repo(org_id, repo_name)

    async def get_branch_protection_rules(self, org_id: str, repo: str) -> list[dict[str, Any]]:
        return await self.graphql_client.get_branch_protection_rules(org_id, repo)

    async def update_branch_protection_rule(
        self,
        org_id: str,
        repo_name: str,
        rule_pattern: str,
        rule_id: str,
        data: dict[str, Any],
    ) -> None:
        await self.graphql_client.update_branch_protection_rule(org_id, repo_name, rule_pattern, rule_id, data)

    async def add_branch_protection_rule(
        self,
        org_id: str,
        repo_name: str,
        repo_node_id: Optional[str],
        data: dict[str, Any],
    ) -> None:
        # in case the repo_id is not available yet, we need to fetch it from GitHub.
        if not repo_node_id:
            repo_data = await self.rest_api.repo.get_repo_data(org_id, repo_name)
            repo_node_id = repo_data["node_id"]

        await self.graphql_client.add_branch_protection_rule(org_id, repo_name, repo_node_id, data)

    async def delete_branch_protection_rule(self, org_id: str, repo_name: str, rule_pattern: str, rule_id: str) -> None:
        await self.graphql_client.delete_branch_protection_rule(org_id, repo_name, rule_pattern, rule_id)

    async def update_repo_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, ruleset: dict[str, Any]) -> None:
        if len(ruleset) > 0:
            await self.rest_api.repo.update_ruleset(org_id, repo_name, ruleset_id, ruleset)

    async def add_repo_ruleset(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        await self.rest_api.repo.add_ruleset(org_id, repo_name, data)

    async def delete_repo_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, name: str) -> None:
        await self.rest_api.repo.delete_ruleset(org_id, repo_name, ruleset_id, name)

    async def get_repo_webhooks(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        return await self.rest_api.repo.get_webhooks(org_id, repo_name)

    async def update_repo_webhook(self, org_id: str, repo_name: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        if len(webhook) > 0:
            await self.rest_api.repo.update_webhook(org_id, repo_name, webhook_id, webhook)

    async def add_repo_webhook(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        await self.rest_api.repo.add_webhook(org_id, repo_name, data)

    async def delete_repo_webhook(self, org_id: str, repo_name: str, webhook_id: int, url: str) -> None:
        await self.rest_api.repo.delete_webhook(org_id, repo_name, webhook_id, url)

    async def get_repo_environments(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        return await self.rest_api.repo.get_environments(org_id, repo_name)

    async def update_repo_environment(self, org_id: str, repo_name: str, env_name: str, env: dict[str, Any]) -> None:
        if len(env) > 0:
            await self.rest_api.repo.update_environment(org_id, repo_name, env_name, env)

    async def add_repo_environment(self, org_id: str, repo_name: str, env_name: str, data: dict[str, str]) -> None:
        await self.rest_api.repo.add_environment(org_id, repo_name, env_name, data)

    async def delete_repo_environment(self, org_id: str, repo_name: str, env_name: str) -> None:
        await self.rest_api.repo.delete_environment(org_id, repo_name, env_name)

    async def get_repo_workflow_settings(self, org_id: str, repo_name: str) -> dict[str, Any]:
        return await self.rest_api.repo.get_workflow_settings(org_id, repo_name)

    async def update_repo_workflow_settings(
        self, org_id: str, repo_name: str, workflow_settings: dict[str, Any]
    ) -> None:
        await self.rest_api.repo.update_workflow_settings(org_id, repo_name, workflow_settings)

    async def get_org_secrets(self, org_id: str) -> list[dict[str, Any]]:
        return await self.rest_api.org.get_secrets(org_id)

    async def update_org_secret(self, org_id: str, secret_name: str, secret: dict[str, Any]) -> None:
        if len(secret) > 0:
            await self.rest_api.org.update_secret(org_id, secret_name, secret)

    async def add_org_secret(self, org_id: str, data: dict[str, str]) -> None:
        await self.rest_api.org.add_secret(org_id, data)

    async def delete_org_secret(self, org_id: str, secret_name: str) -> None:
        await self.rest_api.org.delete_secret(org_id, secret_name)

    async def get_org_variables(self, org_id: str) -> list[dict[str, Any]]:
        return await self.rest_api.org.get_variables(org_id)

    async def update_org_variable(self, org_id: str, variable_name: str, variable: dict[str, Any]) -> None:
        if len(variable) > 0:
            await self.rest_api.org.update_variable(org_id, variable_name, variable)

    async def add_org_variable(self, org_id: str, data: dict[str, str]) -> None:
        await self.rest_api.org.add_variable(org_id, data)

    async def delete_org_variable(self, org_id: str, variable_name: str) -> None:
        await self.rest_api.org.delete_variable(org_id, variable_name)

    async def get_repo_secrets(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        return await self.rest_api.repo.get_secrets(org_id, repo_name)

    async def update_repo_secret(self, org_id: str, repo_name: str, secret_name: str, secret: dict[str, Any]) -> None:
        if len(secret) > 0:
            await self.rest_api.repo.update_secret(org_id, repo_name, secret_name, secret)

    async def add_repo_secret(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        await self.rest_api.repo.add_secret(org_id, repo_name, data)

    async def delete_repo_secret(self, org_id: str, repo_name: str, secret_name: str) -> None:
        await self.rest_api.repo.delete_secret(org_id, repo_name, secret_name)

    async def get_repo_variables(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        return await self.rest_api.repo.get_variables(org_id, repo_name)

    async def update_repo_variable(
        self, org_id: str, repo_name: str, variable_name: str, variable: dict[str, Any]
    ) -> None:
        if len(variable) > 0:
            await self.rest_api.repo.update_variable(org_id, repo_name, variable_name, variable)

    async def add_repo_variable(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        await self.rest_api.repo.add_variable(org_id, repo_name, data)

    async def delete_repo_variable(self, org_id: str, repo_name: str, variable_name: str) -> None:
        await self.rest_api.repo.delete_variable(org_id, repo_name, variable_name)

    async def dispatch_workflow(self, org_id: str, repo_name: str, workflow_name: str) -> bool:
        return await self.rest_api.repo.dispatch_workflow(org_id, repo_name, workflow_name)

    async def get_repo_ids(self, org_id: str, repo_names: list[str]) -> list[str]:
        repo_ids = []
        for repo_name in repo_names:
            repo_data = await self.get_repo_data(org_id, repo_name)
            repo_ids.append(repo_data["id"])
        return repo_ids

    async def get_actor_node_ids(self, actor_names: list[str]) -> list[str]:
        return list(map(lambda x: x[1][1], await self.get_actor_ids_with_type(actor_names)))

    async def get_actor_ids_with_type(self, actor_names: list[str]) -> list[tuple[str, tuple[int, str]]]:
        result = []
        for actor in actor_names:
            if actor.startswith("@"):
                # if it starts with a @, it's either a user or team:
                #    - team-names contains a / in its slug
                #    - user-names are not allowed to contain a /
                if "/" in actor:
                    try:
                        result.append(("Team", await self.rest_api.org.get_team_ids(actor[1:])))
                    except RuntimeError:
                        utils.print_warn(f"team '{actor[1:]}' does not exist, skipping")
                else:
                    try:
                        result.append(("User", await self.rest_api.user.get_user_ids(actor[1:])))
                    except RuntimeError:
                        utils.print_warn(f"user '{actor[1:]}' does not exist, skipping")
            else:
                # it's an app
                try:
                    result.append(("App", await self.rest_api.app.get_app_ids(actor)))
                except RuntimeError:
                    utils.print_warn(f"app '{actor}' does not exist, skipping")

        return result

    async def get_app_node_ids(self, app_names: set[str]) -> dict[str, str]:
        return {app_name: (await self.rest_api.app.get_app_ids(app_name))[1] for app_name in app_names}

    async def get_app_ids(self, app_names: set[str]) -> dict[str, str]:
        return {app_name: (await self.rest_api.app.get_app_ids(app_name))[0] for app_name in app_names}

    async def get_ref_for_pull_request(self, org_id: str, repo_name: str, pull_number: str) -> str:
        return await self.rest_api.repo.get_ref_for_pull_request(org_id, repo_name, pull_number)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Any, Optional

from jsonbender import F, Filter, Forall, If, K, OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.providers.github import GitHubProvider
from otterdog.utils import UNSET, is_set_and_valid, is_unset


@dataclasses.dataclass
class Environment(ModelObject):
    """
    Represents a Deployment Environment of a Repository.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    node_id: str = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    wait_timer: int
    reviewers: list[str]
    deployment_branch_policy: str
    branch_policies: list[str]

    @property
    def model_object_name(self) -> str:
        return "environment"

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if not is_unset(self.wait_timer) and not (0 <= self.wait_timer <= 43200):
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header(parent_object)} has 'wait_timer' of value '{self.wait_timer}' "
                f"outside of supported range (0, 43200).",
            )

        if is_set_and_valid(self.deployment_branch_policy):
            if self.deployment_branch_policy not in {"all", "protected", "selected"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'deployment_branch_policy' of value "
                    f"'{self.deployment_branch_policy}', "
                    f"only values ('all' | 'protected' | 'selected') are allowed.",
                )

            if self.deployment_branch_policy != "selected" and len(self.branch_policies) > 0:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has 'deployment_branch_policy' set to "
                    f"'{self.deployment_branch_policy}', "
                    f"but 'branch_policies' is set to '{self.branch_policies}', "
                    f"setting will be ignored.",
                )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.deployment_branch_policy != "selected":
            if field.name in ["branch_policies"]:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    def include_existing_object_for_live_patch(self, org_id: str, parent_object: Optional[ModelObject]) -> bool:
        from .repository import Repository

        assert isinstance(parent_object, Repository)

        # if it's a repo in the form of "<orgid>.github.io", ignore a missing github-pages environment
        # as it is automatically created, there is a validation rule to warn the user about it.
        if parent_object.name.lower() == f"{org_id}.github.io".lower() and self.name == "github-pages":
            return False
        # if it's a github-pages environment and gh pages are enabled, ignore it.
        # GitHub automatically creates it, a validation warning is output if it is missing
        # in the configuration.
        elif self.name == "github-pages" and parent_object.gh_pages_build_type != "disabled":
            return False
        else:
            return True

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> Environment:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]) -> Environment:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        def transform_reviewers(x):
            match x["type"]:
                case "User":
                    return f'@{x["reviewer"]["login"]}'

                case "Team":
                    return f'@{org_id}/{x["reviewer"]["slug"]}'

                case _:
                    raise RuntimeError("unexpected review type '{x[\"type\"]}'")

        def transform_policy(x):
            if x is None:
                return "all"
            else:
                if x.get("protected_branches", False) is True:
                    return "protected"
                elif x.get("custom_branch_policies", False) is True:
                    return "selected"
                else:
                    raise ValueError(f"unexpected deployment_branch_policy {x}")

        mapping.update(
            {
                "wait_timer": OptionalS("protection_rules", default=[])
                >> Filter(lambda obj: obj.get("type") == "wait_timer")
                >> OptionalS(0, default={})
                >> If(OptionalS("wait_timer") != K(None), S("wait_timer"), K(0)),
                "reviewers": OptionalS("protection_rules", default=[])
                >> Filter(lambda obj: obj.get("type") == "required_reviewers")
                >> OptionalS(0, default={})
                >> OptionalS("reviewers", default=[])
                >> Forall(lambda x: transform_reviewers(x)),
                "deployment_branch_policy": OptionalS("deployment_branch_policy") >> F(lambda x: transform_policy(x)),
                "branch_policies": OptionalS("branch_policies", default=[]) >> Forall(lambda x: x["name"]),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        if "reviewers" in mapping:
            reviewers = data["reviewers"]
            reviewer_mapping = []
            for actor_type, (
                actor_id,
                actor_node_id,
            ) in await provider.get_actor_ids_with_type(reviewers):
                reviewer_mapping.append({"type": actor_type, "id": actor_id})
            mapping["reviewers"] = reviewer_mapping

        if "deployment_branch_policy" in mapping:
            deployment_branch_policy = data["deployment_branch_policy"]

            match deployment_branch_policy:
                case "all":
                    deployment_branch_policy_mapping = None

                case "protected":
                    deployment_branch_policy_mapping = {
                        "protected_branches": True,
                        "custom_branch_policies": False,
                    }

                case "selected":
                    deployment_branch_policy_mapping = {
                        "protected_branches": False,
                        "custom_branch_policies": True,
                    }

                case _:
                    raise RuntimeError(f"unexpected deployment_branch_policy '{deployment_branch_policy}'")

            mapping["deployment_branch_policy"] = deployment_branch_policy_mapping

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return f"orgs.{jsonnet_config.create_environment}"

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        from .repository import Repository

        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, Environment)
                assert isinstance(patch.parent_object, Repository)
                await provider.add_repo_environment(
                    org_id,
                    patch.parent_object.name,
                    patch.expected_object.name,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                assert isinstance(patch.current_object, Environment)
                assert isinstance(patch.parent_object, Repository)
                await provider.delete_repo_environment(org_id, patch.parent_object.name, patch.current_object.name)

            case LivePatchType.CHANGE:
                assert patch.changes is not None
                assert isinstance(patch.current_object, Environment)
                assert isinstance(patch.parent_object, Repository)
                await provider.update_repo_environment(
                    org_id,
                    patch.parent_object.name,
                    patch.current_object.name,
                    await cls.changes_to_provider(org_id, patch.changes, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import is_info_enabled, style

from . import Operation


class ListMembersOperation(Operation):
    """
    Prints statistical information about members of the organization.
    """

    def __init__(self, two_factor_disabled: bool):
        super().__init__()
        self._two_factor_disabled = two_factor_disabled

    @property
    def two_factor_disabled(self) -> bool:
        return self._two_factor_disabled

    @property
    def two_factor_enabled(self) -> bool:
        return self._two_factor_disabled

    def pre_execute(self) -> None:
        if is_info_enabled():
            self.printer.println("Listing organization members:")

    def post_execute(self) -> None:
        pass

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file

            if not os.path.exists(org_file_name):
                self.printer.print_error(
                    f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first"
                )
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
            except RuntimeError as ex:
                self.printer.print_error(f"failed to load configuration: {str(ex)}")
                return 1

            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                members = await provider.rest_api.org.list_members(github_id, self.two_factor_disabled)

            if self.two_factor_disabled is True:
                all_members = await provider.rest_api.org.list_members(github_id, False)
                two_factor_status = (
                    style("enabled", fg="green")
                    if organization.settings.two_factor_requirement is True
                    else style("disabled", fg="red")
                )

                member_status = (
                    style(str(len(members)), fg="green") if len(members) == 0 else style(str(len(members)), fg="red")
                )

                self.printer.println(
                    f"Found {member_status} / {len(all_members)} members with 2FA disabled. "
                    f"Organization has 2FA '{two_factor_status}'"
                )
            else:
                self.printer.println(f"Found {len(members)} members.")

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Any, Iterator, Optional, cast

from jsonbender import F, If, K, OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    UNSET,
    Change,
    IndentingPrinter,
    is_set_and_present,
    is_set_and_valid,
    is_unset,
    write_patch_object_as_json,
)

from .organization_workflow_settings import OrganizationWorkflowSettings


@dataclasses.dataclass
class OrganizationSettings(ModelObject):
    """
    Represents settings of a GitHub Organization.
    """

    name: Optional[str]
    plan: str = dataclasses.field(metadata={"read_only": True})
    description: Optional[str]
    email: Optional[str]
    location: Optional[str]
    company: Optional[str]
    billing_email: str
    twitter_username: Optional[str]
    blog: Optional[str]
    has_discussions: bool
    discussion_source_repository: Optional[str]
    has_organization_projects: bool
    # setting does not seem to be taken into account as the name might suggest,
    # instead has_organization_projects is used to control whether repo can have projects
    # deprecate and schedule for removal
    has_repository_projects: bool
    default_branch_name: str
    default_repository_permission: str
    two_factor_requirement: bool = dataclasses.field(metadata={"read_only": True})
    web_commit_signoff_required: bool
    dependabot_alerts_enabled_for_new_repositories: bool
    dependabot_security_updates_enabled_for_new_repositories: bool
    dependency_graph_enabled_for_new_repositories: bool
    members_can_create_private_repositories: bool
    members_can_create_public_repositories: bool
    members_can_fork_private_repositories: bool
    members_can_create_public_pages: bool
    members_can_change_repo_visibility: bool
    members_can_delete_repositories: bool
    members_can_delete_issues: bool
    members_can_create_teams: bool
    readers_can_create_discussions: bool
    packages_containers_public: bool
    packages_containers_internal: bool
    members_can_change_project_visibility: bool
    security_managers: list[str]

    # nested model fields
    workflows: OrganizationWorkflowSettings = dataclasses.field(metadata={"nested_model": True})

    @property
    def model_object_name(self) -> str:
        return "settings"

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.has_discussions is False and field.name == "discussion_source_repository":
            return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        # execute custom validation rules if present
        self.execute_custom_validation_if_present(context, "validate-org-settings.py")

        # enabling dependabot implicitly enables the dependency graph,
        # disabling the dependency graph in the configuration will result in inconsistencies after
        # applying the configuration, warn the user about it.
        dependabot_alerts_enabled = self.dependabot_alerts_enabled_for_new_repositories is True
        dependabot_security_updates_enabled = self.dependabot_security_updates_enabled_for_new_repositories is True
        dependency_graph_disabled = self.dependency_graph_enabled_for_new_repositories is False

        if is_set_and_present(self.description) and len(self.description) > 160:
            context.add_failure(
                FailureType.ERROR,
                "setting 'description' exceeds maximum allowed length of 160 chars.",
            )

        if (dependabot_alerts_enabled or dependabot_security_updates_enabled) and dependency_graph_disabled:
            context.add_failure(
                FailureType.ERROR,
                "enabling 'dependabot_alerts' or 'dependabot_security_updates' implicitly"
                " enables 'dependency_graph_enabled_for_new_repositories'.",
            )

        if dependabot_security_updates_enabled and not dependabot_alerts_enabled:
            context.add_failure(
                FailureType.ERROR,
                "enabling 'dependabot_security_updates' implicitly enables dependabot_alerts.",
            )

        if self.has_discussions is True and not is_set_and_valid(self.discussion_source_repository):
            context.add_failure(
                FailureType.ERROR,
                "enabling 'has_discussions' requires setting a valid repository in 'discussion_source_repository'.",
            )

        if is_set_and_present(self.discussion_source_repository) and "/" not in self.discussion_source_repository:
            context.add_failure(
                FailureType.ERROR,
                "setting 'discussion_source_repository' requires a repository in '<owner>/<repo-name>' format.",
            )

        if is_set_and_valid(self.default_repository_permission):
            if self.default_repository_permission not in {"none", "read", "write", "admin"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'default_repository_permission' has value '{self.default_repository_permission}', "
                    f"only values ('none' | 'read' | 'write' | 'admin') are allowed.",
                )

        if is_set_and_present(self.workflows):
            self.workflows.validate(context, self)

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject]]:
        if is_set_and_present(self.workflows):
            yield self.workflows, self
            yield from self.workflows.get_model_objects()

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> OrganizationSettings:
        mapping: dict[str, Any] = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        mapping.update(
            {
                "workflows": If(
                    OptionalS("workflows", default=None) == K(None),
                    K(UNSET),
                    S("workflows") >> F(lambda x: OrganizationWorkflowSettings.from_model_data(x)),
                ),
            }
        )

        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]) -> OrganizationSettings:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        mapping.update({"plan": OptionalS("plan", "name", default=UNSET)})
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }
        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return None

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)
        write_patch_object_as_json(patch, printer, False)

        if is_set_and_present(self.workflows):
            default_workflow_settings = cast(OrganizationSettings, default_object).workflows

            patch = self.workflows.get_patch_to(default_workflow_settings)
            if len(patch) > 0:
                printer.print("workflows+:")
                self.workflows.to_jsonnet(printer, jsonnet_config, context, False, default_workflow_settings)

        printer.level_down()
        printer.println("},")

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        assert isinstance(expected_object, OrganizationSettings)
        assert isinstance(current_object, OrganizationSettings)

        modified_settings: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)
        if len(modified_settings) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object, current_object, modified_settings, parent_object, False, cls.apply_live_patch
                )
            )

        context.modified_org_settings = modified_settings

        if is_set_and_valid(expected_object.workflows):
            OrganizationWorkflowSettings.generate_live_patch(
                expected_object.workflows, current_object.workflows, expected_object, context, handler
            )

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        assert patch.patch_type == LivePatchType.CHANGE
        assert patch.changes is not None
        github_settings = await cls.changes_to_provider(org_id, patch.changes, provider)
        await provider.update_org_settings(org_id, github_settings)



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import dataclasses

from quart import render_template

from otterdog.webapp.tasks import Task


@dataclasses.dataclass(repr=False)
class HelpCommentTask(Task[None]):
    installation_id: int
    org_id: str
    repo_name: str
    pull_request_number: int

    async def _pre_execute(self) -> None:
        self.logger.info(
            "adding help text to pull request #%d of repo '%s/%s'",
            self.pull_request_number,
            self.org_id,
            self.repo_name,
        )

    async def _execute(self) -> None:
        rest_api = await self.get_rest_api(self.installation_id)
        comment = await render_template("help_comment.txt")
        await rest_api.issue.create_comment(self.org_id, self.repo_name, str(self.pull_request_number), comment)

    def __repr__(self) -> str:
        return f"HelpCommentTask(repo={self.org_id}/{self.repo_name}, pull_request={self.pull_request_number})"



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any

from otterdog.providers.github.exception import GitHubException
from otterdog.providers.github.rest import RestApi, RestClient
from otterdog.utils import print_debug


class TeamClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_team_slugs(self, org_id: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving teams for org '{org_id}'")

        try:
            response = await self.requester.async_request_json("GET", f"/orgs/{org_id}/teams")
            return list(map(lambda team: team["slug"], response))
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving teams:\n{ex}").with_traceback(tb)

    async def is_user_member_of_team(self, org_id: str, team_slug: str, user: str) -> bool:
        print_debug(f"retrieving membership of user '{user}' for team '{team_slug}' in org '{org_id}'")

        status, body = await self.requester.async_request_raw(
            "GET", f"/orgs/{org_id}/teams/{team_slug}/memberships/{user}"
        )

        if status == 200:
            return True
        elif status == 404:
            return False
        else:
            raise RuntimeError(
                f"failed retrieving team membership for user '{user}' in org '{org_id}'" f"\n{status}: {body}"
            )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from otterdog.config import OrganizationConfig
from otterdog.providers.github import GitHubProvider
from otterdog.utils import style

from . import Operation


class DispatchWorkflowOperation(Operation):
    """
    Dispatches a specified workflow for an organization repo.
    """

    def __init__(self, repo_name: str, workflow_name: str):
        super().__init__()
        self._repo_name = repo_name
        self._workflow_name = workflow_name

    @property
    def repo_name(self) -> str:
        return self._repo_name

    @property
    def workflow_name(self) -> str:
        return self._workflow_name

    def pre_execute(self) -> None:
        self.printer.println(f"Dispatching workflow '{self.workflow_name}' in organization repo '{self.repo_name}':")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                success = await provider.dispatch_workflow(github_id, self.repo_name, self.workflow_name)
                if success is True:
                    self.printer.println(f"workflow '{self.workflow_name}' dispatched for repo '{self.repo_name}'")
                else:
                    self.printer.println(
                        f"failed to dispatch workflow '{self.workflow_name}' for repo '{self.repo_name}'"
                    )

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any, Optional

from otterdog.config import OtterdogConfig
from otterdog.models import LivePatch, ModelObject
from otterdog.models.webhook import Webhook
from otterdog.utils import Change, IndentingPrinter, style

from .diff_operation import DiffOperation, DiffStatus


class PlanOperation(DiffOperation):
    """
    Calculates and displays changes that would be applied to an organization based on the current configuration.
    """

    def __init__(self, no_web_ui: bool, update_webhooks: bool, update_secrets: bool, update_filter: str):
        super().__init__(no_web_ui, update_webhooks, update_secrets, update_filter)

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        super().init(config, printer)

    def pre_execute(self) -> None:
        self.printer.println("Planning execution:")
        self.print_legend()

    def print_legend(self) -> None:
        self.printer.println("\nActions are indicated with the following symbols:")
        self.printer.println(f"  {style('+', fg='green')} create")
        self.printer.println(f"  {style('~', fg='yellow')} modify")
        self.printer.println(f"  {style('!', fg='magenta')} forced update")
        self.printer.println(f"  {style('-', fg='red')} delete")

    def resolve_secrets(self) -> bool:
        return False

    def handle_add_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        self.printer.println()
        model_header = model_object.get_model_header(parent_object)
        self.print_dict(
            model_object.to_model_dict(for_diff=True),
            f"add {model_header}",
            "+",
            "green",
        )

    def handle_delete_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        self.printer.println()
        model_header = model_object.get_model_header(parent_object)
        self.print_dict(
            model_object.to_model_dict(for_diff=True),
            f"remove {model_header}",
            "-",
            "red",
        )

    def handle_modified_object(
        self,
        org_id: str,
        modified_object: dict[str, Change[Any]],
        forced_update: bool,
        current_object: ModelObject,
        expected_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> int:
        self.printer.println()
        model_header = current_object.get_model_header(parent_object)
        self.print_modified_dict(modified_object, model_header, set(), forced_update)

        # FIXME: this code should be moved to the Webhook model class.
        if isinstance(current_object, Webhook):
            if "secret" in modified_object:
                new_secret = modified_object["secret"].to_value
                if not new_secret:
                    self.printer.println(
                        f"\n{style('Warning', fg='red')}: removing secret for webhook "
                        f"with url '{current_object.url}'"
                    )

        settings_to_change = 0
        for k, v in modified_object.items():
            if current_object.is_read_only_key(k):
                self.printer.println(
                    f"\n{style('Note', fg='yellow')}: setting '{k}' " f"is read-only, will be skipped."
                )
            else:
                settings_to_change += 1

        return settings_to_change

    async def handle_finish(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> None:
        self.printer.println(
            f"\n{style('Plan', bright=True)}: {diff_status.additions} to add, "
            f"{diff_status.differences} to change, "
            f"{diff_status.deletions} to delete."
        )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from asyncio import gather
from functools import cached_property
from typing import Any, Iterator

from importlib_resources import files
from playwright.async_api import Error as PlaywrightError
from playwright.async_api import Page, async_playwright

from otterdog import resources, utils
from otterdog.credentials import Credentials


class WebClient:
    # use 10s as default timeout
    _DEFAULT_TIMEOUT = 15000

    def __init__(self, credentials: Credentials):
        self.credentials = credentials

    @cached_property
    def web_settings_definition(self) -> dict[str, Any]:
        # load the definition file which describes how the web settings
        # can be retrieved / modified.
        utils.print_trace("getting web_settings config using jsonnet")

        web_settings_config = files(resources).joinpath("github-web-settings.jsonnet")
        return utils.jsonnet_evaluate_file(str(web_settings_config))

    async def get_org_settings(self, org_id: str, included_keys: set[str]) -> dict[str, Any]:
        utils.print_debug("retrieving settings via web interface")

        async with async_playwright() as playwright:
            try:
                browser = await playwright.firefox.launch()
            except Exception as e:
                tb = e.__traceback__
                raise RuntimeError(
                    "unable to launch browser, make sure you have installed required dependencies using: "
                    "'otterdog install-deps'"
                ).with_traceback(tb) from None

            context = await browser.new_context()

            login_page = await context.new_page()
            login_page.set_default_timeout(self._DEFAULT_TIMEOUT)
            await self._login_if_required(login_page)

            async def process_page(page_url, page_def) -> dict[str, Any]:
                page = await context.new_page()
                page.set_default_timeout(self._DEFAULT_TIMEOUT)
                return await self._retrieve_settings(org_id, page_url, page_def, included_keys, page)

            tasks = [process_page(page_url, page_def) for page_url, page_def in self._get_pages(included_keys)]
            settings_list = await gather(*tasks)
            settings = {k: v for d in settings_list for k, v in d.items()}

            await self._logout(login_page)

            await login_page.close()
            await context.close()
            await browser.close()

            return settings

    def _get_pages(self, included_keys: set[str]) -> Iterator[tuple[str, Any]]:
        for page_url, page_def in self.web_settings_definition.items():
            # check if the page contains any setting that is requested
            if not any(x in included_keys for x in list(map(lambda x: x["name"], page_def))):
                continue
            else:
                yield page_url, page_def

    @staticmethod
    async def _retrieve_settings(
        org_id: str, page_url: str, page_def: Any, included_keys: set[str], page: Page
    ) -> dict[str, Any]:
        settings: dict[str, Any] = {}

        utils.print_trace(f"loading page '{page_url}'")
        response = await page.goto("https://github.com/organizations/{}/{}".format(org_id, page_url))
        assert response is not None
        if not response.ok:
            raise RuntimeError(f"unable to access github page '{page_url}': {response.status}")

        for setting_def in page_def:
            setting = setting_def["name"]
            optional = setting_def["optional"]
            utils.print_trace(f"checking setting '{setting}'")

            if setting not in included_keys:
                continue

            parent = setting_def.get("parent", None)
            if parent is not None:
                parent_value = settings[parent]
                if isinstance(parent_value, bool) and parent_value is False:
                    settings[setting] = None
                    continue

            try:
                setting_type = setting_def["type"]
                match setting_type:
                    case "checkbox":
                        selector = setting_def["selector"]

                    case "radio":
                        selector = f"{setting_def['selector']}:checked"

                    case "select-menu":
                        selector = f"{setting_def['selector']}"

                    case "text":
                        selector = setting_def["selector"]

                    case _:
                        raise RuntimeError(f"not supported setting type '{setting_type}'")

                pre_selector = setting_def["preSelector"]
                if pre_selector is not None:
                    await page.click(pre_selector)
                    await page.wait_for_selector(selector, state="attached")

                value = await page.eval_on_selector(
                    selector,
                    "(el, property) => el[property]",
                    setting_def["valueSelector"],
                )

                if isinstance(value, str):
                    value = value.strip()

                settings[setting] = value
                utils.print_trace(f"retrieved setting for '{setting}' = '{value}'")

            except Exception as e:
                if optional:
                    continue

                if utils.is_debug_enabled():
                    page_name = page_url.split("/")[-1]
                    screenshot_file = f"screenshot_{page_name}.png"
                    await page.screenshot(path=screenshot_file)
                    utils.print_warn(f"saved page screenshot to file '{screenshot_file}'")

                utils.print_warn(f"failed to retrieve setting '{setting}' via web ui:\n{str(e)}")

        return settings

    async def update_org_settings(self, org_id: str, data: dict[str, Any]) -> None:
        utils.print_debug("updating settings via web interface")

        async with async_playwright() as playwright:
            try:
                browser = await playwright.firefox.launch()
            except Exception as e:
                tb = e.__traceback__
                raise RuntimeError(
                    "unable to launch browser, make sure you have installed required dependencies using: "
                    "'otterdog install-deps'"
                ).with_traceback(tb) from None

            page = await browser.new_page()
            page.set_default_timeout(self._DEFAULT_TIMEOUT)

            await self._login_if_required(page)
            await self._update_settings(org_id, data, page)
            await self._logout(page)

            await page.close()
            await browser.close()

            utils.print_debug(f"updated {len(data)} setting(s) via web interface")

    async def _update_settings(self, org_id: str, settings: dict[str, Any], page: Page) -> None:
        # first, collect the set of pages that are need to be loaded
        pages_to_load: dict[str, dict[str, Any]] = {}
        for page_url, page_def in self.web_settings_definition.items():
            for setting_def in page_def:
                setting = setting_def["name"]
                if setting in settings:
                    utils.print_trace(f"adding page '{page_url}' with setting '{setting}'")
                    page_dict = pages_to_load.get(page_url, {})
                    page_dict[setting] = setting_def
                    pages_to_load[page_url] = page_dict

        # second, load the required pages and modify the settings
        for page_url, page_dict in pages_to_load.items():
            utils.print_trace(f"loading page '{page_url}'")
            response = await page.goto("https://github.com/organizations/{}/{}".format(org_id, page_url))
            assert response is not None
            if not response.ok:
                raise RuntimeError(f"unable to access github page '{page_url}': {response.status}")

            for setting, setting_def in page_dict.items():
                utils.print_trace(f"updating setting '{setting}'")
                new_value = settings[setting]

                try:
                    setting_type = setting_def["type"]
                    match setting_type:
                        case "checkbox":
                            await page.set_checked(setting_def["selector"], new_value == "True" or new_value)

                        case "radio":
                            await page.set_checked(f"{setting_def['selector']}[value='{new_value}']", True)

                        case "select-menu":
                            pre_selector = setting_def["preSelector"]
                            await page.click(pre_selector)

                            selector = f"{setting_def['saveSelector']}"
                            await page.wait_for_selector(selector, state="attached")
                            handles = await page.query_selector_all(selector)
                            for handle in handles:
                                if new_value == (await handle.inner_text()).strip():
                                    await handle.click()
                                    break

                        case "text":
                            await page.fill(setting_def["selector"], new_value)

                        case _:
                            raise RuntimeError(f"not supported setting type '{setting_type}'")

                    delay_save = setting_def.get("delay_save", None)
                    if delay_save is not None and delay_save in settings:
                        continue

                    # do a trial run first as this will wait till the button is enabled
                    # this might be needed for some text input forms that perform input validation.
                    await page.click(setting_def["save"], trial=True)
                    await page.click(setting_def["save"], trial=False)

                    utils.print_trace(f"updated setting for '{setting}' = '{new_value}'")
                except Exception as e:
                    if utils.is_debug_enabled():
                        page_name = page_url.split("/")[-1]
                        screenshot_file = f"screenshot_{page_name}.png"
                        await page.screenshot(path=screenshot_file)
                        utils.print_warn(f"saved page screenshot to file '{screenshot_file}'")

                    utils.print_warn(f"failed to update setting '{setting}' via web ui:\n{str(e)}")
                    raise e

    async def open_browser_with_logged_in_user(self, org_id: str) -> None:
        utils.print_debug("opening browser window")

        async with async_playwright() as playwright:
            try:
                browser = await playwright.firefox.launch(headless=False)
            except Exception as e:
                tb = e.__traceback__
                raise RuntimeError(
                    "unable to launch browser, make sure you have installed required dependencies using: "
                    "'otterdog install-deps'"
                ).with_traceback(tb) from None

            context = await browser.new_context(no_viewport=True)

            page = await context.new_page()
            page.set_default_timeout(self._DEFAULT_TIMEOUT)

            await self._login_if_required(page)

            await page.goto("https://github.com/{}".format(org_id))
            input("Enter anything to logout and close browser.\n")

            await self._logout(page)

            await page.close()
            await context.close()
            await browser.close()

    async def _login_if_required(self, page: Page) -> None:
        actor = await self._logged_in_as(page)

        if actor is None:
            await self._login(page)
        elif actor != self.credentials.username:
            raise RuntimeError(f"logged in with unexpected user {actor}")

    @staticmethod
    async def _logged_in_as(page: Page) -> str:
        response = await page.goto("https://github.com/settings/profile")
        assert response is not None
        if not response.ok:
            raise RuntimeError(f"unable to load github profile page: {response.status}")

        try:
            actor = await page.eval_on_selector('meta[name="octolytics-actor-login"]', "element => element.content")
        except PlaywrightError:
            actor = None

        return actor

    async def _login(self, page: Page) -> None:
        response = await page.goto("https://github.com/login")
        assert response is not None
        if not response.ok:
            raise RuntimeError(f"unable to load github login page: {response.status}")

        await page.type("#login_field", self.credentials.username)
        await page.type("#password", self.credentials.password)
        await page.click('input[name="commit"]')

        await page.goto("https://github.com/sessions/two-factor")
        await page.type("#app_totp", self.credentials.totp)

        try:
            actor = await page.eval_on_selector('meta[name="octolytics-actor-login"]', "element => element.content")
            utils.print_trace(f"logged in as {actor}")

            if await page.title() == "Verify two-factor authentication":
                verify_button = page.get_by_role("button", name="Verify 2FA now")
                if verify_button is not None:
                    await verify_button.click()

                    if await page.is_visible('button[text="Confirm"]'):
                        confirm_button = page.get_by_role("button", name="Confirm")
                        if confirm_button is not None:
                            await confirm_button.click()

                    if await page.title() == "Confirm your account recovery settings":
                        confirm_button = page.get_by_role("button", name="Confirm")
                        if confirm_button is not None:
                            await confirm_button.click()

                    await page.type("#app_totp", self.credentials.totp)
        except PlaywrightError as e:
            raise RuntimeError(f"could not log in to web UI: {str(e)}")

    async def _logout(self, page: Page) -> None:
        actor = await self._logged_in_as(page)

        response = await page.goto("https://github.com/logout")
        assert response is not None
        if not response.ok:
            response = await page.goto("https://github.com/settings/profile")
            assert response is not None
            if not response.ok:
                raise RuntimeError("unable to load github logout page")

            try:
                selector = 'summary.Header-link > img[alt = "@{}"]'.format(actor)
                await page.eval_on_selector(selector, "el => el.click()")
                await page.wait_for_selector('button[type="submit"].dropdown-signout')
                await page.eval_on_selector('button[type="submit"].dropdown-signout', "el => el.click()")
            except Exception as e:
                if utils.is_debug_enabled():
                    screenshot_file = "screenshot_profile.png"
                    await page.screenshot(path=screenshot_file)
                    utils.print_warn(f"saved page screenshot to file '{screenshot_file}'")

                raise RuntimeError(f"failed to logout via web ui: {str(e)}")
        else:
            try:
                selector = 'input[value = "Sign out"]'
                await page.eval_on_selector(selector, "el => el.click()")
            except Exception as e:
                if utils.is_debug_enabled():
                    screenshot_file = "screenshot_profile.png"
                    await page.screenshot(path=screenshot_file)
                    utils.print_warn(f"saved page screenshot to file '{screenshot_file}'")

                raise RuntimeError(f"failed to logout via web ui: {str(e)}")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import re
from typing import Any, Optional, cast

from jsonbender import Forall, K, OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    UNSET,
    associate_by_key,
    is_set_and_valid,
    is_unset,
    snake_to_camel_case,
)


@dataclasses.dataclass
class BranchProtectionRule(ModelObject):
    """
    Represents a Branch Protection Rule within a Repository.
    """

    id: str = dataclasses.field(metadata={"external_only": True})
    pattern: str = dataclasses.field(metadata={"key": True})

    allows_deletions: bool
    allows_force_pushes: bool
    is_admin_enforced: bool
    lock_allows_fetch_and_merge: bool
    lock_branch: bool

    requires_pull_request: bool
    # the following settings are only taken into account
    # when requires_pull_request is True
    required_approving_review_count: Optional[int]
    dismisses_stale_reviews: bool
    requires_code_owner_reviews: bool
    require_last_push_approval: bool
    bypass_pull_request_allowances: list[str]
    restricts_review_dismissals: bool
    review_dismissal_allowances: list[str]

    bypass_force_push_allowances: list[str]

    restricts_pushes: bool
    blocks_creations: bool
    push_restrictions: list[str]

    requires_commit_signatures: bool
    requires_conversation_resolution: bool
    requires_linear_history: bool

    requires_status_checks: bool
    requires_strict_status_checks: bool
    required_status_checks: list[str]

    requires_deployments: bool
    required_deployment_environments: list[str]

    @property
    def model_object_name(self) -> str:
        return "branch_protection_rule"

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        # when requires_approving_reviews is false, issue a warning if dependent settings
        # are still set to non default values.

        if self.requires_pull_request is False:
            if is_set_and_valid(self.required_approving_review_count):
                context.add_failure(
                    FailureType.INFO,
                    f"{self.get_model_header(parent_object)} has"
                    f" 'requires_pull_request' disabled but 'required_approving_review_count' "
                    f"is set to '{self.required_approving_review_count}', setting will be ignored.",
                )

            for key in [
                "dismisses_stale_reviews",
                "requires_code_owner_reviews",
                "require_last_push_approval",
                "restricts_review_dismissals",
            ]:
                if self.__getattribute__(key) is True:
                    context.add_failure(
                        FailureType.WARNING,
                        f"{self.get_model_header(parent_object)} has"
                        f" 'requires_pull_request' disabled but '{key}' "
                        f"is enabled, setting will be ignored.",
                    )

            for key in [
                "bypass_pull_request_allowances",
                "review_dismissal_allowances",
            ]:
                value = self.__getattribute__(key)
                if not is_unset(value) and len(value) > 0:
                    context.add_failure(
                        FailureType.WARNING,
                        f"{self.get_model_header(parent_object)} has"
                        f" 'requires_pull_request' disabled but '{key}' "
                        f"is set to '{value}', setting will be ignored.",
                    )

        # required_approving_review_count must be defined when requires_pull_request is enabled
        required_approving_review_count = self.required_approving_review_count
        if self.requires_pull_request is True and not is_unset(required_approving_review_count):
            if required_approving_review_count is None or required_approving_review_count < 0:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has"
                    f" 'requires_pull_request' enabled but 'required_approving_review_count' "
                    f"is not set (must be set to a non negative number).",
                )

        # if 'review_dismissal_allowances' is disabled, issue a warning if review_dismissal_allowances is non-empty.
        review_dismissal_allowances = self.review_dismissal_allowances
        if (
            self.restricts_review_dismissals is False
            and is_set_and_valid(review_dismissal_allowances)
            and len(review_dismissal_allowances) > 0
        ):
            context.add_failure(
                FailureType.INFO,
                f"{self.get_model_header(parent_object)} has"
                f" 'restricts_review_dismissals' disabled but "
                f"'review_dismissal_allowances' is set to '{self.review_dismissal_allowances}', "
                f"setting will be ignored.",
            )

        # if 'allows_force_pushes' is enabled, issue an error if bypass_force_push_allowances is non-empty.
        if (
            self.allows_force_pushes is True
            and is_set_and_valid(self.bypass_force_push_allowances)
            and len(self.bypass_force_push_allowances) > 0
        ):
            context.add_failure(
                FailureType.INFO,
                f"{self.get_model_header(parent_object)} has"
                f" 'allows_force_pushes' enabled but "
                f"'bypass_force_push_allowances' is set to '{self.bypass_force_push_allowances}', "
                f"setting will be ignored.",
            )

        # if 'requires_status_checks' is disabled, issue a warning if required_status_checks is non-empty.
        required_status_checks = self.required_status_checks
        if (
            self.requires_status_checks is False
            and is_set_and_valid(required_status_checks)
            and len(required_status_checks) > 0
        ):
            context.add_failure(
                FailureType.INFO,
                f"{self.get_model_header(parent_object)} has"
                f" 'requires_status_checks' disabled but "
                f"'required_status_checks' is set to '{self.required_status_checks}', "
                f"setting will be ignored.",
            )

        # if 'requires_deployments' is disabled, issue a warning if required_deployment_environments is non-empty.
        if (
            self.requires_deployments is False
            and is_set_and_valid(self.required_deployment_environments)
            and len(self.required_deployment_environments) > 0
        ):
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has "
                f"'requires_deployments' disabled but "
                f"'required_deployment_environments' is set to "
                f"'{self.required_deployment_environments}', setting will be ignored.",
            )

        if self.requires_deployments is True and len(self.required_deployment_environments) > 0:
            from .repository import Repository

            environments = cast(Repository, parent_object).environments

            environments_by_name = associate_by_key(environments, lambda x: x.name)
            for env_name in self.required_deployment_environments:
                if env_name not in environments_by_name:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} requires deployment environment "
                        f"'{env_name}' which is not defined in the repository itself.",
                    )

        # if 'restricts_pushes' is disabled, issue a warning if push_restrictions is non-empty.
        if (
            self.restricts_pushes is False
            and is_set_and_valid(self.push_restrictions)
            and len(self.push_restrictions) > 0
        ):
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has"
                f" 'restricts_pushes' disabled but "
                f"'push_restrictions' is set to '{self.push_restrictions}', "
                f"setting will be ignored.",
            )

        # if 'restricts_pushes' is disabled, issue a warning if blocks_creations is enabled.
        if self.restricts_pushes is False and self.blocks_creations is True:
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has"
                f" 'restricts_pushes' disabled but "
                f"'blocks_creations' is set to '{self.blocks_creations}', "
                f"setting will be ignored.",
            )

        # if 'lock_branch' is disabled, issue a warning if lock_allows_fetch_and_merge is enabled.
        if self.lock_branch is False and self.lock_allows_fetch_and_merge is True:
            context.add_failure(
                FailureType.WARNING,
                f"{self.get_model_header(parent_object)} has 'lock_branch' disabled but "
                f"'lock_allows_fetch_and_merge' enabled, setting will be ignored.",
            )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        # disable diff computation for dependent fields of requires_pull_request,
        if self.requires_pull_request is False:
            if field.name in [
                "required_approving_review_count",
                "dismisses_stale_reviews",
                "requires_code_owner_reviews",
                "require_last_push_approval",
                "bypass_pull_request_allowances",
                "restricts_review_dismissals",
                "review_dismissal_allowances",
            ]:
                return False

        if self.restricts_review_dismissals is False:
            if field.name in ["review_dismissal_allowances"]:
                return False

        if self.allows_force_pushes is True:
            if field.name in ["bypass_force_push_allowances"]:
                return False

        if self.requires_status_checks is False:
            if field.name in [
                "required_status_checks",
                "requires_strict_status_checks",
            ]:
                return False

        if self.requires_deployments is False:
            if field.name in ["required_deployment_environments"]:
                return False

        if self.restricts_pushes is False:
            if field.name in ["push_restrictions"]:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> BranchProtectionRule:
        mapping: dict[str, Any] = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}

        if "requires_approving_reviews" in data:
            mapping["requires_pull_request"] = S("requires_approving_reviews")

        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]) -> BranchProtectionRule:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(snake_to_camel_case(k), default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        mapping["requires_pull_request"] = OptionalS("requiresApprovingReviews", default=UNSET)

        def transform_app(x):
            app = x["app"]
            context = x["context"]

            if app is None:
                app_prefix = "any:"
            else:
                app_slug = app["slug"]
                if app_slug == "github-actions":
                    app_prefix = ""
                else:
                    app_prefix = f"{app_slug}:"

            return f"{app_prefix}{context}"

        mapping["required_status_checks"] = OptionalS("requiredStatusChecks", default=[]) >> Forall(
            lambda x: transform_app(x)
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            snake_to_camel_case(field.name): S(field.name)
            for field in cls.provider_fields()
            if not is_unset(data.get(field.name, UNSET))
        }

        if "requires_pull_request" in data:
            mapping.pop("requiresPullRequest")
            mapping["requiresApprovingReviews"] = K(data["requires_pull_request"])

        if "push_restrictions" in data:
            mapping.pop("pushRestrictions")
            push_restrictions = data["push_restrictions"]
            if is_set_and_valid(push_restrictions):
                actor_ids = await provider.get_actor_node_ids(push_restrictions)
                mapping["pushActorIds"] = K(actor_ids)

        if "review_dismissal_allowances" in data:
            mapping.pop("reviewDismissalAllowances")
            review_dismissal_allowances = data["review_dismissal_allowances"]
            if is_set_and_valid(review_dismissal_allowances):
                actor_ids = await provider.get_actor_node_ids(review_dismissal_allowances)
                mapping["reviewDismissalActorIds"] = K(actor_ids)

        if "bypass_pull_request_allowances" in data:
            mapping.pop("bypassPullRequestAllowances")
            bypass_pull_request_allowances = data["bypass_pull_request_allowances"]
            if is_set_and_valid(bypass_pull_request_allowances):
                actor_ids = await provider.get_actor_node_ids(bypass_pull_request_allowances)
                mapping["bypassPullRequestActorIds"] = K(actor_ids)

        if "bypass_force_push_allowances" in data:
            mapping.pop("bypassForcePushAllowances")
            bypass_force_push_allowances = data["bypass_force_push_allowances"]
            if is_set_and_valid(bypass_force_push_allowances):
                actor_ids = await provider.get_actor_node_ids(bypass_force_push_allowances)
                mapping["bypassForcePushActorIds"] = K(actor_ids)

        if "required_status_checks" in data:
            mapping.pop("requiredStatusChecks")
            required_status_checks = data["required_status_checks"]
            if is_set_and_valid(required_status_checks):
                app_slugs = set()

                for check in required_status_checks:
                    if ":" in check:
                        app_slug, context = re.split(":", check, 1)

                        if app_slug != "any":
                            app_slugs.add(app_slug)
                    else:
                        app_slugs.add("github-actions")

                app_ids = await provider.get_app_node_ids(app_slugs)

                transformed_checks = []
                for check in required_status_checks:
                    if ":" in check:
                        app_slug, context = re.split(":", check, 1)
                    else:
                        app_slug = "github-actions"
                        context = check

                    if app_slug == "any":
                        transformed_checks.append({"appId": "any", "context": context})
                    else:
                        transformed_checks.append({"appId": app_ids[app_slug], "context": context})

                mapping["requiredStatusChecks"] = K(transformed_checks)

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return f"orgs.{jsonnet_config.create_branch_protection_rule}"

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        from .repository import Repository

        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, BranchProtectionRule)
                assert isinstance(patch.parent_object, Repository)
                await provider.add_branch_protection_rule(
                    org_id,
                    patch.parent_object.name,
                    patch.parent_object.node_id,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                assert isinstance(patch.current_object, BranchProtectionRule)
                assert isinstance(patch.parent_object, Repository)
                await provider.delete_branch_protection_rule(
                    org_id, patch.parent_object.name, patch.current_object.pattern, patch.current_object.id
                )

            case LivePatchType.CHANGE:
                assert patch.changes is not None
                assert isinstance(patch.current_object, BranchProtectionRule)
                assert isinstance(patch.parent_object, Repository)
                await provider.update_branch_protection_rule(
                    org_id,
                    patch.parent_object.name,
                    patch.current_object.pattern,
                    patch.current_object.id,
                    await cls.changes_to_provider(org_id, patch.changes, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
from typing import Any, TypeVar

from jsonbender import OptionalS, S, bend  # type: ignore

from otterdog.models import FailureType, ModelObject, ValidationContext
from otterdog.providers.github import GitHubProvider
from otterdog.utils import UNSET, is_unset

VT = TypeVar("VT", bound="Variable")


@dataclasses.dataclass
class Variable(ModelObject, abc.ABC):
    """
    Represents a Variable.
    """

    name: str = dataclasses.field(metadata={"key": True})
    value: str

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if self.name.startswith("GITHUB_"):
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header()} starts with prefix 'GITHUB_' which is not allowed for variables.",
            )

    @classmethod
    def from_model_data(cls, data: dict[str, Any]):
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        return {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import os
import subprocess
from typing import Any

from .utils import (
    jsonnet_evaluate_snippet,
    parse_template_url,
    print_debug,
    print_trace,
    print_warn,
)


class JsonnetConfig:
    # FIXME: the function names to create resources should not be hard-coded but
    #        rather follow a convention to add new resources more easily.

    create_org = "newOrg"
    create_org_webhook = "newOrgWebhook"
    create_org_secret = "newOrgSecret"
    create_org_variable = "newOrgVariable"
    create_repo = "newRepo"
    extend_repo = "extendRepo"
    create_repo_webhook = "newRepoWebhook"
    create_repo_secret = "newRepoSecret"
    create_repo_variable = "newRepoVariable"
    create_branch_protection_rule = "newBranchProtectionRule"
    create_repo_ruleset = "newRepoRuleset"
    create_environment = "newEnvironment"

    def __init__(self, org_id: str, base_dir: str, base_template_url: str, local_only: bool):
        self._org_id = org_id
        self._base_dir = base_dir

        repo_url, file, ref = parse_template_url(base_template_url)

        self._base_template_repo_url = repo_url
        self._base_template_repo_name = os.path.basename(repo_url)
        self._base_template_file = file
        self._base_template_ref = ref

        self._local_only = local_only

        self._default_org_config: dict[str, Any] | None = None
        self._default_org_webhook_config: dict[str, Any] | None = None
        self._default_org_secret_config: dict[str, Any] | None = None
        self._default_org_variable_config: dict[str, Any] | None = None
        self._default_repo_config: dict[str, Any] | None = None
        self._default_repo_webhook_config: dict[str, Any] | None = None
        self._default_repo_secret_config: dict[str, Any] | None = None
        self._default_repo_variable_config: dict[str, Any] | None = None
        self._default_branch_protection_rule_config: dict[str, Any] | None = None
        self._default_repo_ruleset_config: dict[str, Any] | None = None
        self._default_environment_config: dict[str, Any] | None = None

        self._initialized = False

    @property
    def org_id(self) -> str:
        return self._org_id

    def init_template(self) -> None:
        if self._initialized is True:
            return

        if not self._local_only:
            self._init_base_template()

        template_file = self.template_file
        print_debug(f"loading template file '{template_file}'")
        if not os.path.exists(self.template_file):
            raise RuntimeError(f"template file '{template_file}' does not exist")

        # load the default settings for the organization
        self._default_org_config = self.default_org_config_for_org_id("default")

        try:
            # load the default org webhook config
            org_webhook_snippet = f"(import '{template_file}').{self.create_org_webhook}('default')"
            self._default_org_webhook_config = jsonnet_evaluate_snippet(org_webhook_snippet)
        except RuntimeError:
            print_warn("no default org webhook config found, webhooks will be skipped")
            self._default_org_webhook_config = None

        try:
            # load the default org secret config
            org_secret_snippet = f"(import '{template_file}').{self.create_org_secret}('default')"
            self._default_org_secret_config = jsonnet_evaluate_snippet(org_secret_snippet)
        except RuntimeError:
            print_warn("no default org secret config found, secrets will be skipped")
            self._default_org_secret_config = None

        try:
            # load the default org variable config
            org_variable_snippet = f"(import '{template_file}').{self.create_org_variable}('default')"
            self._default_org_variable_config = jsonnet_evaluate_snippet(org_variable_snippet)
        except RuntimeError:
            print_warn("no default org variable config found, variables will be skipped")
            self._default_org_variable_config = None

        try:
            # load the default repo config
            repo_snippet = f"(import '{template_file}').{self.create_repo}('default')"
            self._default_repo_config = jsonnet_evaluate_snippet(repo_snippet)
        except RuntimeError:
            print_warn("no default repo config found, repos will be skipped")
            self._default_repo_config = None

        try:
            # load the default repo webhook config
            repo_webhook_snippet = f"(import '{template_file}').{self.create_repo_webhook}('default')"
            self._default_repo_webhook_config = jsonnet_evaluate_snippet(repo_webhook_snippet)
        except RuntimeError:
            print_warn("no default repo webhook config found, webhooks will be skipped")
            self._default_repo_webhook_config = None

        try:
            # load the default repo secret config
            repo_secret_snippet = f"(import '{template_file}').{self.create_repo_secret}('default')"
            self._default_repo_secret_config = jsonnet_evaluate_snippet(repo_secret_snippet)
        except RuntimeError:
            print_warn("no default repo secret config found, secrets will be skipped")
            self._default_repo_secret_config = None

        try:
            # load the default repo variable config
            repo_variable_snippet = f"(import '{template_file}').{self.create_repo_variable}('default')"
            self._default_repo_variable_config = jsonnet_evaluate_snippet(repo_variable_snippet)
        except RuntimeError:
            print_warn("no default repo variable config found, variables will be skipped")
            self._default_repo_variable_config = None

        try:
            # load the default branch protection rule config
            branch_protection_snippet = f"(import '{template_file}').{self.create_branch_protection_rule}('default')"
            self._default_branch_protection_rule_config = jsonnet_evaluate_snippet(branch_protection_snippet)
        except RuntimeError:
            print_warn("no default branch protection rule config found, branch protection rules will be skipped")
            self._default_branch_protection_rule_config = None

        try:
            # load the default repo ruleset config
            branch_protection_snippet = f"(import '{template_file}').{self.create_repo_ruleset}('default')"
            self._default_repo_ruleset_config = jsonnet_evaluate_snippet(branch_protection_snippet)
        except RuntimeError:
            print_warn("no default repo ruleset config found, rulesets will be skipped")
            self._default_repo_ruleset_config = None

        try:
            # load the default environment config
            environment_snippet = f"(import '{template_file}').{self.create_environment}('default')"
            self._default_environment_config = jsonnet_evaluate_snippet(environment_snippet)
        except RuntimeError:
            print_warn("no default environment config found, environments will be skipped")
            self._default_environment_config = None

        self._initialized = True

    @property
    def default_org_config(self):
        return self._default_org_config

    def default_org_config_for_org_id(self, org_id: str):
        try:
            # load the default settings for the organization
            snippet = f"(import '{self.template_file}').{self.create_org}('{org_id}')"
            return jsonnet_evaluate_snippet(snippet)
        except RuntimeError as ex:
            raise RuntimeError(f"failed to get default organization config for org '{org_id}': {ex}")

    @property
    def default_org_webhook_config(self):
        return self._default_org_webhook_config

    @property
    def default_org_secret_config(self):
        return self._default_org_secret_config

    @property
    def default_org_variable_config(self):
        return self._default_org_variable_config

    @property
    def default_repo_config(self):
        return self._default_repo_config

    @property
    def default_repo_webhook_config(self):
        return self._default_repo_webhook_config

    @property
    def default_repo_secret_config(self):
        return self._default_repo_secret_config

    @property
    def default_repo_variable_config(self):
        return self._default_repo_variable_config

    @property
    def default_branch_protection_rule_config(self):
        return self._default_branch_protection_rule_config

    @property
    def default_repo_ruleset_config(self):
        return self._default_repo_ruleset_config

    @property
    def default_environment_config(self):
        return self._default_environment_config

    @property
    def template_dir(self) -> str:
        return os.path.join(
            self.org_dir,
            "vendor",
            self._base_template_repo_name,
        )

    @property
    def template_file(self) -> str:
        return os.path.join(
            self.template_dir,
            self._base_template_file,
        )

    @property
    def org_dir(self) -> str:
        return f"{self._base_dir}/{self.org_id}"

    @property
    def org_config_file(self) -> str:
        return f"{self.org_dir}/{self.org_id}.jsonnet"

    @property
    def import_statement(self) -> str:
        return f"import 'vendor/{self._base_template_repo_name}/{self._base_template_file}'"

    @property
    def jsonnet_bundle_file(self) -> str:
        return f"{self.org_dir}/jsonnetfile.json"

    @property
    def jsonnet_bundle_lock_file(self) -> str:
        return f"{self.org_dir}/jsonnetfile.lock.json"

    def _init_base_template(self) -> None:
        # create base directory if it does not exist yet
        if not os.path.exists(self.org_dir):
            os.makedirs(self.org_dir)

        content = {
            "version": 1,
            "dependencies": [
                {
                    "source": {
                        "git": {
                            "remote": f"{self._base_template_repo_url}.git",
                            "subdir": "",
                        }
                    },
                    "version": f"{self._base_template_ref}",
                }
            ],
            "legacyImports": True,
        }

        with open(self.jsonnet_bundle_file, "w") as file:
            file.write(json.dumps(content, indent=2))

        # create an empty lock file if it does not exist yet
        lock_file = self.jsonnet_bundle_lock_file
        if not os.path.exists(lock_file):
            with open(lock_file, "w") as file:
                file.write("")

        print_debug("running jsonnet-bundler update")
        cwd = os.getcwd()

        try:
            os.chdir(self.org_dir)
            status, result = subprocess.getstatusoutput("jb update")
            print_trace(f"result = ({status}, {result})")

            if status != 0:
                raise RuntimeError(result)

        finally:
            os.chdir(cwd)

    def __repr__(self) -> str:
        return f"JsonnetConfig('{self._base_dir}, '{self._base_template_file}')"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any

import jq  # type: ignore
from aiohttp.client import ClientSession
from importlib_resources import files

from otterdog import resources
from otterdog.providers.github.auth import AuthStrategy
from otterdog.utils import is_debug_enabled, is_trace_enabled, print_debug, print_trace


class GraphQLClient:
    _GH_GRAPHQL_URL_ROOT = "https://api.github.com/graphql"

    def __init__(self, auth_strategy: AuthStrategy):
        self._auth = auth_strategy.get_auth()

        self._headers = {
            "X-Github-Next-Global-ID": "1",
        }

    async def get_branch_protection_rules(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"async retrieving branch protection rules for repo '{org_id}/{repo_name}'")

        variables = {"organization": org_id, "repository": repo_name}
        branch_protection_rules = await self._async_run_paged_query(variables, "get-branch-protection-rules.gql")

        for branch_protection_rule in branch_protection_rules:
            await self._fill_paged_results_if_not_empty(
                branch_protection_rule,
                "pushAllowances",
                "pushRestrictions",
                "get-push-allowances.gql",
            )

            await self._fill_paged_results_if_not_empty(
                branch_protection_rule,
                "reviewDismissalAllowances",
                "reviewDismissalAllowances",
                "get-review-dismissal-allowances.gql",
            )

            await self._fill_paged_results_if_not_empty(
                branch_protection_rule,
                "bypassPullRequestAllowances",
                "bypassPullRequestAllowances",
                "get-bypass-pull-request-allowances.gql",
            )

            await self._fill_paged_results_if_not_empty(
                branch_protection_rule,
                "bypassForcePushAllowances",
                "bypassForcePushAllowances",
                "get-bypass-force-push-allowances.gql",
            )

        return branch_protection_rules

    async def _fill_paged_results_if_not_empty(
        self,
        branch_protection_rule: dict[str, Any],
        input_key: str,
        output_key: str,
        query_file: str,
    ):
        value = branch_protection_rule.pop(input_key)

        total_count = int(jq.compile(".totalCount").input(value).first())

        if total_count > 0:
            variables = {"branchProtectionRuleId": branch_protection_rule["id"]}
            actors = await self._async_run_paged_query(variables, query_file, f".data.node.{input_key}")
            branch_protection_rule[output_key] = self._transform_actors(actors)
        else:
            branch_protection_rule[output_key] = []

    async def update_branch_protection_rule(
        self,
        org_id: str,
        repo_name: str,
        rule_pattern: str,
        rule_id: str,
        data: dict[str, Any],
    ) -> None:
        print_debug(f"updating branch protection rule '{rule_pattern}' for repo '{org_id}/{repo_name}'")

        data["branchProtectionRuleId"] = rule_id
        variables = {"ruleInput": data}

        query = """mutation($ruleInput: UpdateBranchProtectionRuleInput!) {
           updateBranchProtectionRule(input:$ruleInput) {
             branchProtectionRule {
               pattern
             }
           }
        }"""

        status, body = await self._async_request_raw("POST", query=query, variables=variables)

        if status >= 400:
            raise RuntimeError(f"failed updating branch protection rule '{rule_pattern}' for repo '{repo_name}'")

        json_data = json.loads(body)
        if "data" not in json_data:
            raise RuntimeError(f"failed to update branch protection rule '{rule_pattern}'")

        print_debug(f"successfully updated branch protection rule '{rule_pattern}'")

    async def add_branch_protection_rule(
        self, org_id: str, repo_name: str, repo_node_id: str, data: dict[str, Any]
    ) -> None:
        rule_pattern = data["pattern"]
        print_debug(
            f"creating branch_protection_rule with pattern '{rule_pattern}' " f"for repo '{org_id}/{repo_name}'"
        )

        data["repositoryId"] = repo_node_id
        variables = {"ruleInput": data}

        query = """mutation($ruleInput: CreateBranchProtectionRuleInput!) {
           createBranchProtectionRule(input:$ruleInput) {
             branchProtectionRule {
               pattern
               blocksCreations
             }
           }
        }"""

        status, body = await self._async_request_raw("POST", query, variables)

        if status >= 400:
            raise RuntimeError(f"failed creating branch protection rule '{rule_pattern}' for repo '{repo_name}'")

        json_data = json.loads(body)
        if "data" not in json_data:
            raise RuntimeError(f"failed to create branch protection rule '{rule_pattern}'")

        print_debug(f"successfully created branch protection rule '{rule_pattern}'")

    async def delete_branch_protection_rule(self, org_id: str, repo_name: str, rule_pattern: str, rule_id: str) -> None:
        print_debug(f"deleting branch protection rule '{rule_pattern}' for repo '{org_id}/{repo_name}'")

        variables = {"ruleInput": {"branchProtectionRuleId": rule_id}}

        query = """mutation($ruleInput: DeleteBranchProtectionRuleInput!) {
           deleteBranchProtectionRule(input:$ruleInput) {
             clientMutationId
           }
        }"""

        status, body = await self._async_request_raw("POST", query, variables)

        if status >= 400:
            raise RuntimeError(f"failed removing branch protection rule '{rule_pattern}' for repo '{repo_name}'")

        print_debug(f"successfully removed branch protection rule '{rule_pattern}'")

    async def _async_run_paged_query(
        self,
        input_variables: dict[str, Any],
        query_file: str,
        prefix_selector: str = ".data.repository.branchProtectionRules",
    ) -> list[dict[str, Any]]:
        print_debug(f"running async graphql query '{query_file}' with input '{json.dumps(input_variables)}'")

        query = files(resources).joinpath(f"graphql/{query_file}").read_text()

        finished = False
        end_cursor = None
        result = []

        while not finished:
            variables = {"endCursor": end_cursor}
            variables.update(input_variables)

            async with ClientSession() as session:
                headers = self._headers.copy()
                self._auth.update_headers_with_authorization(headers)

                async with session.post(
                    url=f"{self._GH_GRAPHQL_URL_ROOT}",
                    headers=headers,
                    json={"query": query, "variables": variables},
                ) as response:
                    if is_debug_enabled():
                        print_debug(
                            f"graphql query '{query_file}' with input '{json.dumps(input_variables)}': "
                            f"rate-limit-used = {response.headers.get('x-ratelimit-used', None)}"
                        )

                    if is_trace_enabled():
                        print_trace(f"graphql result = ({response.status}, {await response.text()})")

                    if not response.ok:
                        raise RuntimeError(f"failed running query '{query_file}'")

                    json_data = await response.json()

            if "data" in json_data:
                rules_result = jq.compile(prefix_selector + ".nodes").input(json_data).first()

                for rule in rules_result:
                    result.append(rule)

                page_info = jq.compile(prefix_selector + ".pageInfo").input(json_data).first()

                if page_info["hasNextPage"]:
                    end_cursor = page_info["endCursor"]
                else:
                    finished = True
            else:
                raise RuntimeError(f"failed running graphql query '{query_file}'")

        return result

    async def _async_request_raw(self, method: str, query: str, variables: dict[str, Any]) -> tuple[int, str]:
        print_trace(f"async '{method}', query = {query}, variables = {variables}")

        headers = self._headers.copy()
        self._auth.update_headers_with_authorization(headers)

        async with ClientSession() as session:
            async with session.request(
                method,
                url=self._GH_GRAPHQL_URL_ROOT,
                headers=headers,
                json={"query": query, "variables": variables},
            ) as response:
                text = await response.text()
                status = response.status

                if is_trace_enabled():
                    print_trace(f"async '{method}' result = ({status}, {text})")

                return status, text

    @staticmethod
    def _transform_actors(actors: list[dict[str, Any]]) -> list[str]:
        result = []
        for actor_wrapper in actors:
            actor = actor_wrapper["actor"]
            typename = actor["__typename"]
            if typename == "User":
                login = actor["login"]
                result.append(f"@{login}")
            elif typename == "Team":
                slug = actor["combinedSlug"]
                result.append(f"@{slug}")
            elif typename == "App":
                slug = actor["slug"]
                result.append(slug)
            else:
                raise RuntimeError(f"unsupported actor '{actor}'")

        return result



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import filecmp
import os
import re
from io import StringIO
from tempfile import TemporaryDirectory
from typing import Union, cast

from pydantic import ValidationError
from quart import current_app, render_template

from otterdog.config import OrganizationConfig
from otterdog.models import LivePatch
from otterdog.operations.diff_operation import DiffStatus
from otterdog.operations.local_plan import LocalPlanOperation
from otterdog.providers.github import RestApi
from otterdog.utils import IndentingPrinter, LogLevel
from otterdog.webapp.tasks import Task, get_otterdog_config
from otterdog.webapp.webhook.github_models import PullRequest, Repository


@dataclasses.dataclass
class ValidationResult:
    plan_output: str = ""
    validation_success: bool = True
    requires_secrets: bool = False


@dataclasses.dataclass(repr=False)
class ValidatePullRequestTask(Task[ValidationResult]):
    """Validates a PR and adds the result as a comment."""

    installation_id: int
    org_id: str
    repository: Repository
    pull_request_or_number: PullRequest | int
    log_level: LogLevel = LogLevel.WARN

    @property
    def check_base_config(self) -> bool:
        return True

    async def _pre_execute(self) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        if isinstance(self.pull_request_or_number, int):
            response = await rest_api.pull_request.get_pull_request(
                self.org_id, self.repository.name, str(self.pull_request_or_number)
            )
            try:
                self.pull_request = PullRequest.model_validate(response)
            except ValidationError as ex:
                self.logger.exception("failed to load pull request event data", exc_info=ex)
                return
        else:
            self.pull_request = cast(PullRequest, self.pull_request_or_number)

        self.logger.info(
            "validating pull request #%d of repo '%s' with log level '%s'",
            self.pull_request.number,
            self.repository.full_name,
            self.log_level,
        )

        await self._create_pending_status(rest_api)

        from .check_sync import CheckConfigurationInSyncTask

        check_task = CheckConfigurationInSyncTask(
            self.installation_id,
            self.org_id,
            self.repository,
            self.pull_request_or_number,
        )

        current_app.add_background_task(check_task.execute)

    async def _post_execute(self, result_or_exception: Union[ValidationResult, Exception]) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        if isinstance(result_or_exception, Exception):
            await self._create_failure_status(rest_api)
        else:
            await self._update_final_status(rest_api, result_or_exception)

    async def _execute(self) -> ValidationResult:
        otterdog_config = get_otterdog_config()
        pull_request_number = str(self.pull_request.number)
        project_name = otterdog_config.get_project_name(self.org_id) or self.org_id

        rest_api = await self.get_rest_api(self.installation_id)

        with TemporaryDirectory(dir=otterdog_config.jsonnet_base_dir) as work_dir:
            org_config = OrganizationConfig.of(
                project_name,
                self.org_id,
                {"provider": "inmemory", "api_token": rest_api.token},
                work_dir,
                otterdog_config,
            )

            jsonnet_config = org_config.jsonnet_config
            if not os.path.exists(jsonnet_config.org_dir):
                os.makedirs(jsonnet_config.org_dir)

            jsonnet_config.init_template()

            # get BASE config
            base_file = jsonnet_config.org_config_file + "-BASE"
            await get_config(
                rest_api,
                self.org_id,
                self.org_id,
                otterdog_config.default_config_repo,
                base_file,
                self.pull_request.base.ref,
            )

            # get HEAD config from PR
            head_file = jsonnet_config.org_config_file
            await get_config(
                rest_api,
                self.org_id,
                self.pull_request.head.repo.owner.login,
                self.pull_request.head.repo.name,
                head_file,
                self.pull_request.head.ref,
            )

            validation_result = ValidationResult()

            if filecmp.cmp(base_file, head_file):
                self.logger.debug("head and base config are identical, no need to validate")
                validation_result.plan_output = "No changes."
                validation_result.validation_success = True
            else:
                output = StringIO()
                printer = IndentingPrinter(output, log_level=self.log_level)
                operation = LocalPlanOperation("-BASE", False, False, "")

                def callback(org_id: str, diff_status: DiffStatus, patches: list[LivePatch]):
                    validation_result.requires_secrets = any(list(map(lambda x: x.requires_secrets(), patches)))

                operation.set_callback(callback)
                operation.init(otterdog_config, printer)

                plan_result = await operation.execute(org_config)

                validation_result.plan_output = output.getvalue()
                validation_result.validation_success = plan_result == 0
                self.logger.info("local plan:" + validation_result.plan_output)

            warnings = []
            if validation_result.requires_secrets:
                warnings.append("some of requested changes require secrets, need to apply these changes manually")

            comment = await render_template(
                "validation_comment.txt",
                sha=self.pull_request.head.sha,
                result=escape_for_github(validation_result.plan_output),
                warnings=warnings,
                admin_team=f"{self.org_id}/{get_admin_team()}",
            )

            # add a comment about the validation result to the PR
            await rest_api.issue.create_comment(
                self.org_id, otterdog_config.default_config_repo, pull_request_number, comment
            )

            return validation_result

    async def _create_pending_status(self, rest_api: RestApi):
        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            "pending",
            _get_webhook_validation_context(),
            "validating configuration change using otterdog",
        )

    async def _create_failure_status(self, rest_api: RestApi):
        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            "failure",
            _get_webhook_validation_context(),
            "otterdog validation failed, please contact an admin",
        )

    async def _update_final_status(self, rest_api: RestApi, validation_result: ValidationResult) -> None:
        if validation_result.validation_success is True:
            desc = "otterdog validation completed successfully"
            status = "success"
        else:
            desc = "otterdog validation failed, check validation result in comment history"
            status = "error"

        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            status,
            _get_webhook_validation_context(),
            desc,
        )

    def __repr__(self) -> str:
        pull_request_number = (
            self.pull_request_or_number
            if isinstance(self.pull_request_or_number, int)
            else self.pull_request_or_number.number
        )
        return f"ValidatePullRequestTask(repo={self.repository.full_name}, pull_request={pull_request_number})"


def _get_webhook_validation_context() -> str:
    return current_app.config["GITHUB_WEBHOOK_VALIDATION_CONTEXT"]


def get_admin_team() -> str:
    return current_app.config["GITHUB_ADMIN_TEAM"]


async def get_config(rest_api: RestApi, org_id: str, owner: str, repo: str, filename: str, ref: str):
    path = f"otterdog/{org_id}.jsonnet"
    content = await rest_api.content.get_content(
        owner,
        repo,
        path,
        ref,
    )
    with open(filename, "w") as file:
        file.write(content)


def escape_for_github(text: str) -> str:
    lines = text.splitlines()

    output = []
    for line in lines:
        ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]")
        line = ansi_escape.sub("", line)

        diff_escape = re.compile(r"(\s+)([-+!])(\s+)")
        line = diff_escape.sub(r"\g<2>\g<1>", line)

        diff_escape2 = re.compile(r"(\s+)(~)")
        line = diff_escape2.sub(r"!\g<1>", line)

        output.append(line)

    return "\n".join(output)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Any, Optional, cast

from jsonbender import Forall, If, K, OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import FailureType, LivePatch, LivePatchType, ValidationContext
from otterdog.models.variable import Variable
from otterdog.providers.github import GitHubProvider
from otterdog.utils import UNSET, is_set_and_valid, is_unset


@dataclasses.dataclass
class OrganizationVariable(Variable):
    """
    Represents a Variable defined on organization level.
    """

    visibility: str
    selected_repositories: list[str]

    @property
    def model_object_name(self) -> str:
        return "org_variable"

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        super().validate(context, parent_object)

        if is_set_and_valid(self.visibility):
            from .github_organization import GitHubOrganization

            org = cast(GitHubOrganization, parent_object)
            if self.visibility == "private" and org.settings.plan == "free":
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'visibility' of value "
                    f"'{self.visibility}' which is not available for organization with free plan.",
                )
            elif self.visibility not in {"public", "private", "selected"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'visibility' of value "
                    f"'{self.visibility}', "
                    f"only values ('public' | 'private' | 'selected') are allowed.",
                )

            if self.visibility != "selected" and len(self.selected_repositories) > 0:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has 'visibility' set to "
                    f"'{self.visibility}', "
                    f"but 'selected_repositories' is set to '{self.selected_repositories}', "
                    f"setting will be ignored.",
                )

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        mapping.update(
            {
                "visibility": If(
                    S("visibility") == K("all"),
                    K("public"),
                    OptionalS("visibility", default=UNSET),
                ),
                "selected_repositories": OptionalS("selected_repositories", default=[]) >> Forall(lambda x: x["name"]),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping: dict[str, Any] = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        if "visibility" in mapping:
            mapping["visibility"] = If(S("visibility") == K("public"), K("all"), S("visibility"))

        if "selected_repositories" in mapping:
            mapping.pop("selected_repositories")
            mapping["selected_repository_ids"] = K(await provider.get_repo_ids(org_id, data["selected_repositories"]))

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return f"orgs.{jsonnet_config.create_org_variable}"

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, OrganizationVariable)
                await provider.add_org_variable(
                    org_id,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                assert isinstance(patch.current_object, OrganizationVariable)
                await provider.delete_org_variable(org_id, patch.current_object.name)

            case LivePatchType.CHANGE:
                assert isinstance(patch.expected_object, OrganizationVariable)
                assert isinstance(patch.current_object, OrganizationVariable)
                await provider.update_org_variable(
                    org_id,
                    patch.current_object.name,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

# Original code from https://github.com/go-build-it/quart-github-webhook
# licensed under Apache License Version 2.0

import collections
import hashlib
import hmac
import json
import logging

from quart import abort, request


class GitHubWebhook(object):
    def __init__(self):
        self._endpoint = None
        self._secret = None
        self._hooks = collections.defaultdict(list)
        self._logger = logging.getLogger(__name__)

    def init_app(self, app) -> None:
        """
        Initializes this webhook on the given :code:`app`.
        """

        self._endpoint = app.config["GITHUB_WEBHOOK_ENDPOINT"]

        secret = app.config["GITHUB_WEBHOOK_SECRET"]
        if secret is not None and not isinstance(secret, bytes):
            secret = secret.encode("utf-8")
        self._secret = secret

        app.add_url_rule(
            rule=self._endpoint,
            endpoint=self._endpoint,
            view_func=self._post_receive,
            methods=["POST"],
        )

    def hook(self, event_type):
        """
        Registers a function as a hook. Multiple hooks can be registered for a given type, but the
        order in which they are invoke is unspecified.

        :param event_type: The event type this hook will be invoked for.
        """

        def decorator(func):
            self._hooks[event_type].append(func)
            return func

        return decorator

    async def _get_digest(self):
        """Return message digest if a secret key was provided"""

        if self._secret:
            return hmac.new(self._secret, await request.data, hashlib.sha1).hexdigest()

    async def _post_receive(self):
        """Callback from Flask"""

        digest = await self._get_digest()

        if digest is not None:
            sig_parts = _get_header("X-Hub-Signature").split("=", 1)
            if not isinstance(digest, str):
                digest = str(digest)

            if len(sig_parts) < 2 or sig_parts[0] != "sha1" or not hmac.compare_digest(sig_parts[1], digest):
                abort(400, "Invalid signature")

        event_type = _get_header("X-Github-Event")
        content_type = _get_header("content-type")
        if content_type == "application/x-www-form-urlencoded":
            formdata = (await request.form).to_dict(flat=True)
            data = json.loads(formdata["payload"])
        elif content_type == "application/json":
            data = await request.get_json()
        else:
            abort(415, f"Unknown content type {content_type}")

        if data is None:
            abort(400, "Request body must contain data")

        self._logger.info(
            "%s (%s)",
            _format_event(event_type, data),
            _get_header("X-Github-Delivery"),
        )

        for hook in self._hooks.get(event_type, []):
            await hook(data)

        return "", 204


def _get_header(key):
    """Return message header"""

    try:
        return request.headers[key]
    except KeyError:
        abort(400, "Missing header: " + key)


EVENT_DESCRIPTIONS = {
    "commit_comment": "{comment[user][login]} commented on " "{comment[commit_id]} in {repository[full_name]}",
    "create": "{sender[login]} created {ref_type} ({ref}) in " "{repository[full_name]}",
    "delete": "{sender[login]} deleted {ref_type} ({ref}) in " "{repository[full_name]}",
    "deployment": "{sender[login]} deployed {deployment[ref]} to "
    "{deployment[environment]} in {repository[full_name]}",
    "deployment_status": "deployment of {deployement[ref]} to "
    "{deployment[environment]} "
    "{deployment_status[state]} in "
    "{repository[full_name]}",
    "fork": "{forkee[owner][login]} forked {forkee[name]}",
    "gollum": "{sender[login]} edited wiki pages in {repository[full_name]}",
    "issue_comment": "{sender[login]} commented on issue #{issue[number]} " "in {repository[full_name]}",
    "issues": "{sender[login]} {action} issue #{issue[number]} in " "{repository[full_name]}",
    "member": "{sender[login]} {action} member {member[login]} in " "{repository[full_name]}",
    "membership": "{sender[login]} {action} member {member[login]} to team "
    "{team[name]} in "
    "{repository[full_name]}",
    "page_build": "{sender[login]} built pages in {repository[full_name]}",
    "ping": "ping from {sender[login]}",
    "public": "{sender[login]} publicized {repository[full_name]}",
    "pull_request": "{sender[login]} {action} pull #{pull_request[number]} in " "{repository[full_name]}",
    "pull_request_review": "{sender[login]} {action} {review[state]} "
    "review on pull #{pull_request[number]} in "
    "{repository[full_name]}",
    "pull_request_review_comment": "{comment[user][login]} {action} comment "
    "on pull #{pull_request[number]} in "
    "{repository[full_name]}",
    "push": "{pusher[name]} pushed {ref} in {repository[full_name]}",
    "release": "{release[author][login]} {action} {release[tag_name]} in " "{repository[full_name]}",
    "repository": "{sender[login]} {action} repository " "{repository[full_name]}",
    "status": "{sender[login]} set {sha} status to {state} in " "{repository[full_name]}",
    "team_add": "{sender[login]} added repository {repository[full_name]} to " "team {team[name]}",
    "watch": "{sender[login]} {action} watch in repository " "{repository[full_name]}",
}


def _format_event(event_type, data):
    try:
        return EVENT_DESCRIPTIONS[event_type].format(**data)
    except KeyError:
        return event_type



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************
import logging
import os
from sys import exit

from decouple import config  # type: ignore

from otterdog.webapp import create_app
from otterdog.webapp.config import config_dict

# WARNING: Don't run with debug turned on in production!
DEBUG: bool = config("DEBUG", default=True, cast=bool)

# Determine which configuration to use
config_mode = "Debug" if DEBUG else "Production"

try:
    app_config = config_dict[config_mode]
except KeyError:
    exit("Error: Invalid <config_mode>. Expected values [Debug, Production] ")

app = create_app(app_config)  # type: ignore

logging.basicConfig(level=logging.INFO)

if os.path.exists(app_config.APP_ROOT):
    os.chdir(app_config.APP_ROOT)
else:
    app.logger.error(f"APP_ROOT '{app_config.APP_ROOT}' does not exist, exiting.")
    exit(1)

if DEBUG:
    app.logger.info("DEBUG       = " + str(DEBUG))
    app.logger.info("Environment = " + config_mode)
    app.logger.info("QUART_APP   = " + app_config.QUART_APP)
    app.logger.info("APP_ROOT    = " + app_config.APP_ROOT)


def run():
    app.run(debug=True)


if __name__ == "__main__":
    run()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os.path
from typing import Optional

from otterdog.config import OrganizationConfig
from otterdog.providers.github import GitHubProvider
from otterdog.utils import style

from . import Operation


class PushOperation(Operation):
    """
    Pushes a local configuration of an organization to its meta-data repository.
    """

    def __init__(self, push_message: Optional[str]):
        super().__init__()
        self._push_message = push_message

    @property
    def push_message(self) -> Optional[str]:
        return self._push_message

    def pre_execute(self) -> None:
        self.printer.println("Pushing organization configurations:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")

        org_file_name = jsonnet_config.org_config_file

        if not os.path.exists(org_file_name):
            self.printer.print_error(
                f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first"
            )
            return 1

        self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with open(org_file_name, "r") as file:
                content = file.read()

            with open(jsonnet_config.jsonnet_bundle_file, "r") as file:
                bundle_content = file.read()

            with GitHubProvider(credentials) as provider:
                try:
                    updated_files = []
                    updated = False

                    if await provider.update_content(
                        org_config.github_id,
                        org_config.config_repo,
                        f"otterdog/{github_id}.jsonnet",
                        content,
                        self.push_message,
                    ):
                        updated_files.append(f"otterdog/{github_id}.jsonnet")
                        updated = True

                    if await provider.update_content(
                        org_config.github_id,
                        org_config.config_repo,
                        "otterdog/jsonnetfile.json",
                        bundle_content,
                        self.push_message,
                    ):
                        updated_files.append("otterdog/jsonnetfile.json")
                        updated |= True

                    if await provider.update_content(
                        org_config.github_id,
                        org_config.config_repo,
                        "otterdog/jsonnetfile.lock.json",
                        "",
                        self.push_message,
                    ):
                        updated_files.append("otterdog/jsonnetfile.lock.json")
                        updated |= True

                except RuntimeError as e:
                    self.printer.print_error(
                        f"failed to push definition to repo '{org_config.github_id}/{org_config.config_repo}': {str(e)}"
                    )
                    return 1

            if updated:
                self.printer.println(
                    f"organization definition pushed to repo '{org_config.github_id}/{org_config.config_repo}': "
                )
                for updated_file in updated_files:
                    self.printer.println(f"  - '{updated_file}'")
            else:
                self.printer.println("no changes, nothing pushed")

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import dataclasses
import os
from io import StringIO
from tempfile import TemporaryDirectory

from quart import render_template

from otterdog.config import OrganizationConfig
from otterdog.operations.apply import ApplyOperation
from otterdog.utils import IndentingPrinter, LogLevel
from otterdog.webapp.tasks import Task, get_otterdog_config
from otterdog.webapp.webhook.github_models import PullRequest, Repository

from .validate_pull_request import escape_for_github, get_config


@dataclasses.dataclass(repr=False)
class ApplyChangesTask(Task[None]):
    """Applies changes from a merged PR and adds the result as a comment."""

    installation_id: int
    org_id: str
    repository: Repository
    pull_request: PullRequest

    async def _execute(self) -> None:
        if self.pull_request.base.ref != self.repository.default_branch:
            self.logger.debug(
                "pull request merged into '%s' which is not the default branch '%s', ignoring",
                self.pull_request.base.ref,
                self.repository.default_branch,
            )
            return

        assert self.pull_request.merged is True
        assert self.pull_request.merge_commit_sha is not None

        self.logger.info(
            "applying merged pull request #%d of repo '%s'", self.pull_request.number, self.repository.full_name
        )

        otterdog_config = get_otterdog_config()
        project_name = otterdog_config.get_project_name(self.org_id) or self.org_id
        pull_request_number = str(self.pull_request.number)

        rest_api = await self.get_rest_api(self.installation_id)

        with TemporaryDirectory(dir=otterdog_config.jsonnet_base_dir) as work_dir:
            org_config = OrganizationConfig.of(
                project_name,
                self.org_id,
                {"provider": "inmemory", "api_token": rest_api.token},
                work_dir,
                otterdog_config,
            )

            jsonnet_config = org_config.jsonnet_config

            if not os.path.exists(jsonnet_config.org_dir):
                os.makedirs(jsonnet_config.org_dir)

            jsonnet_config.init_template()

            # get config from merge commit sha
            head_file = jsonnet_config.org_config_file
            await get_config(
                rest_api,
                self.org_id,
                self.org_id,
                otterdog_config.default_config_repo,
                head_file,
                self.pull_request.merge_commit_sha,
            )

            output = StringIO()
            printer = IndentingPrinter(output, log_level=LogLevel.ERROR)

            # let's create an apply operation that forces processing but does not update
            # any web UI settings and resources using credentials
            operation = ApplyOperation(
                force_processing=True,
                no_web_ui=True,
                update_webhooks=False,
                update_secrets=False,
                update_filter="",
                delete_resources=True,
                resolve_secrets=False,
                include_resources_with_secrets=False,
            )
            operation.init(otterdog_config, printer)

            await operation.execute(org_config)

            text = output.getvalue()
            self.logger.info(text)

            result = await render_template("applied_changes_comment.txt", result=escape_for_github(text))

            await rest_api.issue.create_comment(
                self.org_id, otterdog_config.default_config_repo, pull_request_number, result
            )

    def __repr__(self) -> str:
        return f"ApplyChangesTask(repo={self.repository.full_name}, pull_request={self.pull_request.number})"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Any, Optional

from jsonbender import Forall, If, K, S  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.models.workflow_settings import WorkflowSettings
from otterdog.providers.github import GitHubProvider
from otterdog.utils import Change, is_set_and_valid


@dataclasses.dataclass
class OrganizationWorkflowSettings(WorkflowSettings):
    """
    Represents workflow settings defined on organization level.
    """

    enabled_repositories: str
    selected_repositories: list[str]

    @property
    def model_object_name(self) -> str:
        return "org_workflow_settings"

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.enabled_repositories == "none":
            if field.name == "enabled_repositories":
                return True
            else:
                return False

        if field.name == "selected_repositories":
            if self.enabled_repositories == "selected":
                return True
            else:
                return False

        return super().include_field_for_diff_computation(field)

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        super().validate(context, parent_object)

        if is_set_and_valid(self.enabled_repositories):
            if self.enabled_repositories not in {"all", "none", "selected"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'enabled_repositories' of value "
                    f"'{self.enabled_repositories}', "
                    f"only values ('all' | 'none' | 'selected') are allowed.",
                )

            if self.enabled_repositories != "selected" and len(self.selected_repositories) > 0:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has 'enabled_repositories' set to "
                    f"'{self.enabled_repositories}', "
                    f"but 'selected_repositories' is set to '{self.selected_repositories}', "
                    f"setting will be ignored.",
                )

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)
        mapping.update(
            {
                "selected_repositories": If(
                    S("selected_repositories") == K(None),
                    K([]),
                    S("selected_repositories") >> Forall(lambda x: x["name"]),
                ),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = await super().get_mapping_to_provider(org_id, data, provider)

        if "selected_repositories" in data:
            mapping.pop("selected_repositories")
            mapping["selected_repository_ids"] = K(await provider.get_repo_ids(org_id, data["selected_repositories"]))

        return mapping

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        assert isinstance(expected_object, OrganizationWorkflowSettings)
        assert isinstance(current_object, OrganizationWorkflowSettings)

        modified_workflow_settings: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)

        # FIXME: needed to add this hack to ensure that enabled_repositories is also present in
        #        the modified data as GitHub has made this property required.
        if "allowed_actions" in modified_workflow_settings:
            enabled_repositories = expected_object.enabled_repositories
            modified_workflow_settings["enabled_repositories"] = Change(enabled_repositories, enabled_repositories)

        if len(modified_workflow_settings) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_workflow_settings,
                    parent_object,
                    False,
                    cls.apply_live_patch,
                )
            )

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        assert patch.patch_type == LivePatchType.CHANGE
        assert patch.changes is not None
        github_settings = await cls.changes_to_provider(org_id, patch.changes, provider)
        await provider.update_org_workflow_settings(org_id, github_settings)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Any, Optional, cast

from jsonbender import S  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.models.organization_settings import OrganizationSettings
from otterdog.models.organization_workflow_settings import OrganizationWorkflowSettings
from otterdog.models.workflow_settings import WorkflowSettings
from otterdog.providers.github import GitHubProvider
from otterdog.utils import UNSET, Change, is_set_and_valid, is_unset


@dataclasses.dataclass
class RepositoryWorkflowSettings(WorkflowSettings):
    """
    Represents workflow settings defined on repository level.
    """

    enabled: bool

    @property
    def model_object_name(self) -> str:
        return "repo_workflow_settings"

    def coerce_from_org_settings(
        self, parent_object: Optional[ModelObject], org_workflow_settings: OrganizationWorkflowSettings
    ) -> RepositoryWorkflowSettings:
        copy = dataclasses.replace(self)

        if org_workflow_settings.enabled_repositories == "none":
            copy.enabled = UNSET  # type: ignore

        from otterdog.models.repository import Repository

        repository_name = cast(Repository, parent_object).name

        if (
            org_workflow_settings.enabled_repositories == "selected"
            and repository_name not in org_workflow_settings.selected_repositories
        ):
            copy.enabled = UNSET  # type: ignore

        if org_workflow_settings.are_actions_more_restricted(self.allowed_actions):
            copy.allowed_actions = UNSET  # type: ignore
            for prop in self._selected_action_properties:
                copy.__setattr__(prop, UNSET)

        if org_workflow_settings.default_workflow_permissions == "read":
            copy.default_workflow_permissions = UNSET  # type: ignore

        if org_workflow_settings.actions_can_approve_pull_request_reviews is False:
            copy.actions_can_approve_pull_request_reviews = UNSET  # type: ignore

        return copy

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        super().validate(context, parent_object)

        if is_set_and_valid(self.enabled) and self.enabled is True:
            from .github_organization import GitHubOrganization

            org_workflow_settings = cast(GitHubOrganization, context.root_object).settings.workflows

            if org_workflow_settings.enabled_repositories == "none" and self.enabled is True:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has enabled workflows, "
                    f"while on organization level it disabled for all repositories, setting will be ignored.",
                )

            from otterdog.models.repository import Repository

            repository_name = cast(Repository, parent_object).name

            if (
                org_workflow_settings.enabled_repositories == "selected"
                and repository_name not in org_workflow_settings.selected_repositories
                and self.enabled is True
            ):
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has enabled workflows, "
                    f"while on organization level it is only enabled for selected repositories, "
                    f"setting will be ignored.",
                )

            if (
                org_workflow_settings.default_workflow_permissions == "read"
                and self.default_workflow_permissions == "write"
            ):
                context.add_failure(
                    FailureType.INFO,
                    f"{self.get_model_header(parent_object)} has 'default_workflow_permissions' of value "
                    f"'{self.default_workflow_permissions}', "
                    f"while on organization level it is restricted to "
                    f"'{org_workflow_settings.default_workflow_permissions}', setting will be ignored.",
                )

            if (
                org_workflow_settings.actions_can_approve_pull_request_reviews is False
                and self.actions_can_approve_pull_request_reviews is True
            ):
                context.add_failure(
                    FailureType.INFO,
                    f"{self.get_model_header(parent_object)} has 'actions_can_approve_pull_request_reviews' "
                    "enabled, while on organization level it is disabled, setting will be ignored.",
                )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if is_unset(self.enabled):
            return False

        if self.enabled is False:
            if field.name == "enabled":
                return True
            else:
                return False

        return super().include_field_for_diff_computation(field)

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        if "enabled" in data and data["enabled"] is False:
            return {"enabled": S("enabled")}
        else:
            return await super().get_mapping_to_provider(org_id, data, provider)

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: Optional[ModelObject],
        current_object: Optional[ModelObject],
        parent_object: Optional[ModelObject],
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        assert isinstance(expected_object, RepositoryWorkflowSettings)

        expected_org_settings = cast(OrganizationSettings, context.expected_org_settings)
        coerced_object = expected_object.coerce_from_org_settings(parent_object, expected_org_settings.workflows)

        if current_object is None:
            handler(LivePatch.of_addition(coerced_object, parent_object, coerced_object.apply_live_patch))
            return

        assert isinstance(current_object, RepositoryWorkflowSettings)

        modified_workflow_settings: dict[str, Change[Any]] = coerced_object.get_difference_from(current_object)

        # FIXME: needed to add this hack to ensure that enabled is also present in
        #        the modified data as GitHub has made this property required.
        if len(modified_workflow_settings) > 0:
            if coerced_object.enabled is True and expected_org_settings.workflows.allowed_actions == "local_only":
                modified_workflow_settings["allowed_actions"] = Change(current_object.allowed_actions, "local_only")

            if "allowed_actions" in modified_workflow_settings:
                modified_workflow_settings["enabled"] = Change(current_object.enabled, coerced_object.enabled)

        if len(modified_workflow_settings) > 0:
            handler(
                LivePatch.of_changes(
                    coerced_object,
                    current_object,
                    modified_workflow_settings,
                    parent_object,
                    False,
                    cls.apply_live_patch,
                )
            )

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        from .repository import Repository

        assert isinstance(patch.parent_object, Repository)

        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, RepositoryWorkflowSettings)
                await provider.update_repo_workflow_settings(
                    org_id,
                    patch.parent_object.name,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.CHANGE:
                assert patch.changes is not None
                github_settings = await cls.changes_to_provider(org_id, patch.changes, provider)
                await provider.update_repo_workflow_settings(org_id, patch.parent_object.name, github_settings)

            case _:
                raise RuntimeError(f"unexpected patch type '{patch.patch_type}'")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import dataclasses
import time
from abc import abstractmethod
from typing import Any, Optional, Protocol

import mintotp  # type: ignore

from otterdog import utils


@dataclasses.dataclass
class Credentials:
    """
    A simple data class to hold credential information to access GitHub.
    """

    _username: Optional[str]
    _password: Optional[str]
    _totp_secret: Optional[str]
    _github_token: Optional[str]

    _last_totp: Optional[str] = None

    @property
    def username(self) -> str:
        if self._username is None:
            raise RuntimeError("username not available")
        else:
            return self._username

    @property
    def password(self) -> str:
        if self._password is None:
            raise RuntimeError("password not available")
        else:
            return self._password

    @property
    def totp(self) -> str:
        if self._totp_secret is None:
            raise RuntimeError("totp_secret not available")

        while True:
            totp = mintotp.totp(self._totp_secret)
            utils.print_trace(f"generated totp '{totp}'")

            if self._last_totp is None or totp != self._last_totp:
                self._last_totp = totp
                return totp
            else:
                utils.print_info("waiting 3s till generating new totp ...")
                time.sleep(3)

    @property
    def github_token(self) -> str:
        if self._github_token is None:
            raise RuntimeError("github_token not available")
        else:
            return self._github_token

    def __str__(self) -> str:
        return "Credentials(username={})".format(self.username)


class CredentialProvider(Protocol):
    @abstractmethod
    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials:
        ...

    @abstractmethod
    def get_secret(self, data: str) -> str:
        ...



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from otterdog.config import OrganizationConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import style

from . import Operation


class ShowLiveOperation(Operation):
    """
    Shows the current live configuration of organizations.
    """

    def __init__(self, no_web_ui: bool):
        super().__init__()
        self._no_web_ui = no_web_ui

    @property
    def no_web_ui(self) -> bool:
        return self._no_web_ui

    def pre_execute(self) -> None:
        self.printer.println("Showing live resources:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
        self.printer.level_up()

        try:
            try:
                credentials = self.config.get_credentials(org_config)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{str(e)}")
                return 1

            with GitHubProvider(credentials) as provider:
                if self.no_web_ui is True:
                    self.printer.print_warn(
                        "the Web UI will not be queried as '--no-web-ui' has been specified, "
                        "the resulting config will be incomplete."
                    )

                organization = await GitHubOrganization.load_from_provider(
                    github_id, jsonnet_config, provider, self.no_web_ui, self.printer
                )

            for model_object, parent_object in organization.get_model_objects():
                self.printer.println()
                model_header = model_object.get_model_header(parent_object)
                self.print_dict(model_object.to_model_dict(), model_header, "", "black")

            return 0

        finally:
            self.printer.level_down()
